from metaflow import metaflow_config
from metaflow.exception import MetaflowException

MAGIC_FILE = "conda.manifest"


# TODO: This can be lifted all the way into metaflow config
def _datastore_packageroot(datastore, echo):
    datastore_type = datastore.TYPE
    datastore_packageroot = getattr(
        metaflow_config,
        "CONDA_PACKAGE_{datastore_type}ROOT".format(
            datastore_type=datastore_type.upper()
        ),
        None,
    )
    if datastore_packageroot is None:
        datastore_sysroot = datastore.get_datastore_root_from_config(echo)
        if datastore_sysroot is None:
            # TODO: Throw a more evocative error message
            raise MetaflowException(
                msg="METAFLOW_DATASTORE_SYSROOT_{datastore_type} must be set!".format(
                    datastore_type=datastore_type.upper()
                )
            )
        datastore_packageroot = "{datastore_sysroot}/conda".format(
            datastore_sysroot=datastore_sysroot
        )
    return datastore_packageroot



from metaflow.decorators import FlowDecorator


# TODO (savin): Lift this decorator up since it's also used by Argo now
class ScheduleDecorator(FlowDecorator):
    """
    Specifies the times when the flow should be run when running on a
    production scheduler.

    Parameters
    ----------
    hourly : bool, default False
        Run the workflow hourly.
    daily : bool, default True
        Run the workflow daily.
    weekly : bool, default False
        Run the workflow weekly.
    cron : str, optional, default None
        Run the workflow at [a custom Cron schedule](https://docs.aws.amazon.com/eventbridge/latest/userguide/scheduled-events.html#cron-expressions)
        specified by this expression.
    timezone : str, optional, default None
        Timezone on which the schedule runs (default: None). Currently supported only for Argo workflows,
        which accepts timezones in [IANA format](https://nodatime.org/TimeZones).
    """

    name = "schedule"
    defaults = {
        "cron": None,
        "weekly": False,
        "daily": True,
        "hourly": False,
        "timezone": None,
    }

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        # Currently supports quartz cron expressions in UTC as defined in
        # https://docs.aws.amazon.com/eventbridge/latest/userguide/scheduled-events.html#cron-expressions
        if self.attributes["cron"]:
            self.schedule = self.attributes["cron"]
        elif self.attributes["weekly"]:
            self.schedule = "0 0 ? * SUN *"
        elif self.attributes["hourly"]:
            self.schedule = "0 * * * ? *"
        elif self.attributes["daily"]:
            self.schedule = "0 0 * * ? *"
        else:
            self.schedule = None

        # Argo Workflows supports the IANA timezone standard, e.g. America/Los_Angeles
        self.timezone = self.attributes["timezone"]



import os
import inspect
import importlib
import sys


def main(flow_file, fn_name_or_path, run_pathspec):
    hook_fn = None

    try:
        module_path, function_name = fn_name_or_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        hook_fn = getattr(module, function_name)
    except (ImportError, AttributeError, ValueError):
        try:
            module_name = os.path.splitext(os.path.basename(flow_file))[0]
            spec = importlib.util.spec_from_file_location(module_name, flow_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            hook_fn = getattr(module, fn_name_or_path)
        except (AttributeError, IOError) as e:
            print(
                f"[exit_hook] Could not load function '{fn_name_or_path}' "
                f"as an import path or from '{flow_file}': {e}"
            )
            sys.exit(1)

    argspec = inspect.getfullargspec(hook_fn)

    # Check if fn expects a run object as an arg.
    if "run" in argspec.args or argspec.varkw is not None:
        from metaflow import Run

        try:
            _run = Run(run_pathspec, _namespace_check=False)
        except Exception as ex:
            print(ex)
            _run = None

        hook_fn(run=_run)
    else:
        hook_fn()


if __name__ == "__main__":
    try:
        flow_file, fn_name, run_pathspec = sys.argv[1:4]
    except Exception:
        print("Usage: exit_hook_script.py <flow_file> <function_name> <run_pathspec>")
        sys.exit(1)

    main(flow_file, fn_name, run_pathspec)



import base64
import json
import re
from hashlib import sha1

from metaflow import JSONType, current, decorators, parameters
from metaflow._vendor import click
from metaflow.exception import MetaflowException, MetaflowInternalError
from metaflow.metaflow_config import (
    FEAT_ALWAYS_UPLOAD_CODE_PACKAGE,
    SERVICE_VERSION_CHECK,
    SFN_STATE_MACHINE_PREFIX,
    SFN_COMPRESS_STATE_MACHINE,
    UI_URL,
)
from metaflow.package import MetaflowPackage
from metaflow.plugins.aws.batch.batch_decorator import BatchDecorator
from metaflow.tagging_util import validate_tags
from metaflow.util import get_username, to_bytes, to_unicode, version_parse

from .production_token import load_token, new_token, store_token
from .step_functions import StepFunctions
from metaflow.tagging_util import validate_tags
from ..aws_utils import validate_aws_tag

VALID_NAME = re.compile(r"[^a-zA-Z0-9_\-\.]")


class IncorrectProductionToken(MetaflowException):
    headline = "Incorrect production token"


class RunIdMismatch(MetaflowException):
    headline = "Run ID mismatch"


class IncorrectMetadataServiceVersion(MetaflowException):
    headline = "Incorrect version for metaflow service"


class StepFunctionsStateMachineNameTooLong(MetaflowException):
    headline = "AWS Step Functions state machine name too long"


@click.group()
def cli():
    pass


@cli.group(help="Commands related to AWS Step Functions.")
@click.option(
    "--name",
    default=None,
    type=str,
    help="State Machine name. The flow name is used instead "
    "if this option is not specified",
)
@click.pass_obj
def step_functions(obj, name=None):
    obj.check(obj.graph, obj.flow, obj.environment, pylint=obj.pylint)
    (
        obj.state_machine_name,
        obj.token_prefix,
        obj.is_project,
    ) = resolve_state_machine_name(obj, name)


@step_functions.command(
    help="Deploy a new version of this workflow to " "AWS Step Functions."
)
@click.option(
    "--authorize",
    default=None,
    help="Authorize using this production token. You need this "
    "when you are re-deploying an existing flow for the first "
    "time. The token is cached in METAFLOW_HOME, so you only "
    "need to specify this once.",
)
@click.option(
    "--generate-new-token",
    is_flag=True,
    help="Generate a new production token for this flow. "
    "This will move the production flow to a new "
    "namespace.",
)
@click.option(
    "--new-token",
    "given_token",
    default=None,
    help="Use the given production token for this flow. "
    "This will move the production flow to the given "
    "namespace.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    default=None,
    help="Annotate all objects produced by AWS Step Functions runs "
    "with the given tag. You can specify this option multiple "
    "times to attach multiple tags.",
)
@click.option(
    "--aws-batch-tag",
    "aws_batch_tags",
    multiple=True,
    default=None,
    help="AWS Batch tags.",
)
@click.option(
    "--namespace",
    "user_namespace",
    default=None,
    help="Change the namespace from the default (production token) "
    "to the given tag. See run --help for more information.",
)
@click.option(
    "--only-json",
    is_flag=True,
    default=False,
    help="Only print out JSON sent to AWS Step Functions. Do not " "deploy anything.",
)
@click.option(
    "--max-workers",
    default=100,
    show_default=True,
    help="Maximum number of parallel processes.",
)
@click.option(
    "--workflow-timeout", default=None, type=int, help="Workflow timeout in seconds."
)
@click.option(
    "--log-execution-history",
    is_flag=True,
    help="Log AWS Step Functions execution history to AWS CloudWatch "
    "Logs log group.",
)
@click.option(
    "--use-distributed-map/--no-use-distributed-map",
    is_flag=True,
    help="Use AWS Step Functions Distributed Map instead of Inline Map for "
    "defining foreach tasks in Amazon State Language.",
)
@click.option(
    "--compress-state-machine/--no-compress-state-machine",
    is_flag=True,
    default=SFN_COMPRESS_STATE_MACHINE,
    help="Compress AWS Step Functions state machine to fit within the 8K limit.",
)
@click.option(
    "--deployer-attribute-file",
    default=None,
    show_default=True,
    type=str,
    help="Write the workflow name to the file specified. Used internally for Metaflow's Deployer API.",
    hidden=True,
)
@click.pass_obj
def create(
    obj,
    tags=None,
    aws_batch_tags=None,
    user_namespace=None,
    only_json=False,
    authorize=None,
    generate_new_token=False,
    given_token=None,
    max_workers=None,
    workflow_timeout=None,
    log_execution_history=False,
    use_distributed_map=False,
    compress_state_machine=False,
    deployer_attribute_file=None,
):
    for node in obj.graph:
        if any([d.name == "slurm" for d in node.decorators]):
            raise MetaflowException(
                "Step *%s* is marked for execution on Slurm with AWS Step Functions which isn't currently supported."
                % node.name
            )

    validate_tags(tags)

    if deployer_attribute_file:
        with open(deployer_attribute_file, "w") as f:
            json.dump(
                {
                    "name": obj.state_machine_name,
                    "flow_name": obj.flow.name,
                    "metadata": obj.metadata.metadata_str(),
                },
                f,
            )

    obj.echo(
        "Deploying *%s* to AWS Step Functions..." % obj.state_machine_name, bold=True
    )

    if SERVICE_VERSION_CHECK:
        check_metadata_service_version(obj)

    token = resolve_token(
        obj.state_machine_name,
        obj.token_prefix,
        obj,
        authorize,
        given_token,
        generate_new_token,
        obj.is_project,
    )

    flow = make_flow(
        obj,
        token,
        obj.state_machine_name,
        tags,
        aws_batch_tags,
        user_namespace,
        max_workers,
        workflow_timeout,
        obj.is_project,
        use_distributed_map,
        compress_state_machine,
    )

    if only_json:
        obj.echo_always(flow.to_json(), err=False, no_bold=True)
    else:
        flow.deploy(log_execution_history)
        obj.echo(
            "State Machine *{state_machine}* "
            "for flow *{name}* pushed to "
            "AWS Step Functions successfully.\n".format(
                state_machine=obj.state_machine_name, name=current.flow_name
            ),
            bold=True,
        )
        if obj._is_state_machine_name_hashed:
            obj.echo(
                "Note that the flow was deployed with a truncated name "
                "due to a length limit on AWS Step Functions. The "
                "original long name is stored in task metadata.\n"
            )
        flow.schedule()
        obj.echo("What will trigger execution of the workflow:", bold=True)
        obj.echo(flow.trigger_explanation(), indent=True)


def check_metadata_service_version(obj):
    metadata = obj.metadata
    version = metadata.version()
    if version == "local":
        return
    elif version is not None and version_parse(version) >= version_parse("2.0.2"):
        # Metaflow metadata service needs to be at least at version 2.0.2
        return
    else:
        obj.echo("")
        obj.echo(
            "You are running a version of the metaflow service "
            "that currently doesn't support AWS Step Functions. "
        )
        obj.echo(
            "For more information on how to upgrade your "
            "service to a compatible version (>= 2.0.2), visit:"
        )
        obj.echo(
            "    https://docs.outerbounds.com/engineering/operations/migration/",
            fg="green",
        )
        obj.echo(
            "Once you have upgraded your metadata service, please "
            "re-execute your command."
        )
        raise IncorrectMetadataServiceVersion(
            "Try again with a more recent " "version of metaflow service " "(>=2.0.2)."
        )


def resolve_state_machine_name(obj, name):
    def attach_prefix(name: str):
        if SFN_STATE_MACHINE_PREFIX is not None and (
            not name.startswith(SFN_STATE_MACHINE_PREFIX)
        ):
            return SFN_STATE_MACHINE_PREFIX + "_" + name
        return name

    project = current.get("project_name")
    obj._is_state_machine_name_hashed = False
    if project:
        if name:
            raise MetaflowException(
                "--name is not supported for @projects. " "Use --branch instead."
            )
        state_machine_name = attach_prefix(current.project_flow_name)
        project_branch = to_bytes(".".join((project, current.branch_name)))
        token_prefix = (
            "mfprj-%s"
            % to_unicode(base64.b32encode(sha1(project_branch).digest()))[:16]
        )
        is_project = True
        # AWS Step Functions has a limit of 80 chars for state machine names.
        # We truncate the state machine name if the computed name is greater
        # than 60 chars and append a hashed suffix to ensure uniqueness.
        if len(state_machine_name) > 60:
            name_hash = to_unicode(
                base64.b32encode(sha1(to_bytes(state_machine_name)).digest())
            )[:16].lower()
            state_machine_name = "%s-%s" % (state_machine_name[:60], name_hash)
            obj._is_state_machine_name_hashed = True
    else:
        if name and VALID_NAME.search(name):
            raise MetaflowException("Name '%s' contains invalid characters." % name)

        state_machine_name = attach_prefix(name if name else current.flow_name)
        token_prefix = state_machine_name
        is_project = False

        if len(state_machine_name) > 80:
            msg = (
                "The full name of the workflow:\n*%s*\nis longer than 80 "
                "characters.\n\n"
                "To deploy this workflow to AWS Step Functions, please "
                "assign a shorter name\nusing the option\n"
                "*step-functions --name <name> create*." % state_machine_name
            )
            raise StepFunctionsStateMachineNameTooLong(msg)

    return state_machine_name, token_prefix.lower(), is_project


def make_flow(
    obj,
    token,
    name,
    tags,
    aws_batch_tags,
    namespace,
    max_workers,
    workflow_timeout,
    is_project,
    use_distributed_map,
    compress_state_machine=False,
):
    if obj.flow_datastore.TYPE != "s3":
        raise MetaflowException("AWS Step Functions requires --datastore=s3.")

    # Attach AWS Batch decorator to the flow
    decorators._attach_decorators(obj.flow, [BatchDecorator.name])
    decorators._process_late_attached_decorator(
        [BatchDecorator.name],
        obj.flow,
        obj.graph,
        obj.environment,
        obj.flow_datastore,
        obj.logger,
    )
    obj.graph = obj.flow._graph

    obj.package = MetaflowPackage(
        obj.flow,
        obj.environment,
        obj.echo,
        suffixes=obj.package_suffixes,
        flow_datastore=obj.flow_datastore if FEAT_ALWAYS_UPLOAD_CODE_PACKAGE else None,
    )
    # This blocks until the package is created
    if FEAT_ALWAYS_UPLOAD_CODE_PACKAGE:
        package_url = obj.package.package_url()
        package_sha = obj.package.package_sha()
    else:
        package_url, package_sha = obj.flow_datastore.save_data(
            [obj.package.blob], len_hint=1
        )[0]

    if aws_batch_tags is not None:
        batch_tags = {}
        for item in list(aws_batch_tags):
            key, value = item.split("=")
            # These are fresh AWS tags provided by the user through the CLI,
            # so we must validate them.
            validate_aws_tag(key, value)
            batch_tags[key] = value

    return StepFunctions(
        name,
        obj.graph,
        obj.flow,
        obj.package.package_metadata,
        package_sha,
        package_url,
        token,
        obj.metadata,
        obj.flow_datastore,
        obj.environment,
        obj.event_logger,
        obj.monitor,
        tags=tags,
        aws_batch_tags=batch_tags,
        namespace=namespace,
        max_workers=max_workers,
        username=get_username(),
        workflow_timeout=workflow_timeout,
        is_project=is_project,
        use_distributed_map=use_distributed_map,
        compress_state_machine=compress_state_machine,
    )


def resolve_token(
    name, token_prefix, obj, authorize, given_token, generate_new_token, is_project
):
    # 1) retrieve the previous deployment, if one exists
    workflow = StepFunctions.get_existing_deployment(name)
    if workflow is None:
        obj.echo(
            "It seems this is the first time you are deploying *%s* to "
            "AWS Step Functions." % name
        )
        prev_token = None
    else:
        prev_user, prev_token = workflow

    # 2) authorize this deployment
    if prev_token is not None:
        if authorize is None:
            authorize = load_token(token_prefix)
        elif authorize.startswith("production:"):
            authorize = authorize[11:]

        # we allow the user who deployed the previous version to re-deploy,
        # even if they don't have the token
        if prev_user != get_username() and authorize != prev_token:
            obj.echo(
                "There is an existing version of *%s* on AWS Step "
                "Functions which was deployed by the user "
                "*%s*." % (name, prev_user)
            )
            obj.echo(
                "To deploy a new version of this flow, you need to use "
                "the same production token that they used. "
            )
            obj.echo(
                "Please reach out to them to get the token. Once you "
                "have it, call this command:"
            )
            obj.echo("    step-functions create --authorize MY_TOKEN", fg="green")
            obj.echo(
                'See "Organizing Results" at docs.metaflow.org for more '
                "information about production tokens."
            )
            raise IncorrectProductionToken(
                "Try again with the correct " "production token."
            )

    # 3) do we need a new token or should we use the existing token?
    if given_token:
        if is_project:
            # we rely on a known prefix for @project tokens, so we can't
            # allow the user to specify a custom token with an arbitrary prefix
            raise MetaflowException(
                "--new-token is not supported for "
                "@projects. Use --generate-new-token to "
                "create a new token."
            )
        if given_token.startswith("production:"):
            given_token = given_token[11:]
        token = given_token
        obj.echo("")
        obj.echo("Using the given token, *%s*." % token)
    elif prev_token is None or generate_new_token:
        token = new_token(token_prefix, prev_token)
        if token is None:
            if prev_token is None:
                raise MetaflowInternalError(
                    "We could not generate a new " "token. This is unexpected. "
                )
            else:
                raise MetaflowException(
                    "--generate-new-token option is not "
                    "supported after using --new-token. "
                    "Use --new-token to make a new "
                    "namespace."
                )
        obj.echo("")
        obj.echo("A new production token generated.")
    else:
        token = prev_token

    obj.echo("")
    obj.echo("The namespace of this production flow is")
    obj.echo("    production:%s" % token, fg="green")
    obj.echo(
        "To analyze results of this production flow " "add this line in your notebooks:"
    )
    obj.echo('    namespace("production:%s")' % token, fg="green")
    obj.echo(
        "If you want to authorize other people to deploy new versions "
        "of this flow to AWS Step Functions, they need to call"
    )
    obj.echo("    step-functions create --authorize %s" % token, fg="green")
    obj.echo("when deploying this flow to AWS Step Functions for the first " "time.")
    obj.echo(
        'See "Organizing Results" at https://docs.metaflow.org/ for more '
        "information about production tokens."
    )
    obj.echo("")
    store_token(token_prefix, token)
    return token


@parameters.add_custom_parameters(deploy_mode=False)
@step_functions.command(help="Trigger the workflow on AWS Step Functions.")
@click.option(
    "--run-id-file",
    default=None,
    show_default=True,
    type=str,
    help="Write the ID of this run to the file specified.",
)
@click.option(
    "--deployer-attribute-file",
    default=None,
    show_default=True,
    type=str,
    help="Write the metadata and pathspec of this run to the file specified.\nUsed internally for Metaflow's Deployer API.",
    hidden=True,
)
@click.pass_obj
def trigger(obj, run_id_file=None, deployer_attribute_file=None, **kwargs):
    def _convert_value(param):
        # Swap `-` with `_` in parameter name to match click's behavior
        val = kwargs.get(param.name.replace("-", "_").lower())
        if param.kwargs.get("type") == JSONType:
            val = json.dumps(val)
        elif isinstance(val, parameters.DelayedEvaluationParameter):
            val = val(return_str=True)
        return val

    params = {
        param.name: _convert_value(param)
        for _, param in obj.flow._get_parameters()
        if kwargs.get(param.name.replace("-", "_").lower()) is not None
    }

    response = StepFunctions.trigger(obj.state_machine_name, params)

    id = response["executionArn"].split(":")[-1]
    run_id = "sfn-" + id

    if run_id_file:
        with open(run_id_file, "w") as f:
            f.write(str(run_id))

    if deployer_attribute_file:
        with open(deployer_attribute_file, "w") as f:
            json.dump(
                {
                    "name": obj.state_machine_name,
                    "metadata": obj.metadata.metadata_str(),
                    "pathspec": "/".join((obj.flow.name, run_id)),
                },
                f,
            )

    obj.echo(
        "Workflow *{name}* triggered on AWS Step Functions "
        "(run-id *{run_id}*).".format(name=obj.state_machine_name, run_id=run_id),
        bold=True,
    )

    run_url = (
        "%s/%s/%s" % (UI_URL.rstrip("/"), obj.flow.name, run_id) if UI_URL else None
    )

    if run_url:
        obj.echo(
            "See the run in the UI at %s" % run_url,
            bold=True,
        )


@step_functions.command(help="List all runs of the workflow on AWS Step Functions.")
@click.option(
    "--running",
    default=False,
    is_flag=True,
    help="List all runs of the workflow in RUNNING state on " "AWS Step Functions.",
)
@click.option(
    "--succeeded",
    default=False,
    is_flag=True,
    help="List all runs of the workflow in SUCCEEDED state on " "AWS Step Functions.",
)
@click.option(
    "--failed",
    default=False,
    is_flag=True,
    help="List all runs of the workflow in FAILED state on " "AWS Step Functions.",
)
@click.option(
    "--timed-out",
    default=False,
    is_flag=True,
    help="List all runs of the workflow in TIMED_OUT state on " "AWS Step Functions.",
)
@click.option(
    "--aborted",
    default=False,
    is_flag=True,
    help="List all runs of the workflow in ABORTED state on " "AWS Step Functions.",
)
@click.pass_obj
def list_runs(
    obj, running=False, succeeded=False, failed=False, timed_out=False, aborted=False
):
    states = []
    if running:
        states.append("RUNNING")
    if succeeded:
        states.append("SUCCEEDED")
    if failed:
        states.append("FAILED")
    if timed_out:
        states.append("TIMED_OUT")
    if aborted:
        states.append("ABORTED")
    executions = StepFunctions.list(obj.state_machine_name, states)
    found = False
    for execution in executions:
        found = True
        if execution.get("stopDate"):
            obj.echo(
                "*sfn-{id}* "
                "startedAt:'{startDate}' "
                "stoppedAt:'{stopDate}' "
                "*{status}*".format(
                    id=execution["name"],
                    status=execution["status"],
                    startDate=execution["startDate"].replace(microsecond=0),
                    stopDate=execution["stopDate"].replace(microsecond=0),
                )
            )
        else:
            obj.echo(
                "*sfn-{id}* "
                "startedAt:'{startDate}' "
                "*{status}*".format(
                    id=execution["name"],
                    status=execution["status"],
                    startDate=execution["startDate"].replace(microsecond=0),
                )
            )
    if not found:
        if len(states) > 0:
            status = ""
            for idx, state in enumerate(states):
                if idx == 0:
                    pass
                elif idx == len(states) - 1:
                    status += " and "
                else:
                    status += ", "
                status += "*%s*" % state
            obj.echo(
                "No %s executions for *%s* found on AWS Step Functions."
                % (status, obj.state_machine_name)
            )
        else:
            obj.echo(
                "No executions for *%s* found on AWS Step Functions."
                % (obj.state_machine_name)
            )


@step_functions.command(help="Delete a workflow")
@click.option(
    "--authorize",
    default=None,
    type=str,
    help="Authorize the deletion with a production token",
)
@click.pass_obj
def delete(obj, authorize=None):
    def _token_instructions(flow_name, prev_user):
        obj.echo(
            "There is an existing version of *%s* on AWS Step "
            "Functions which was deployed by the user "
            "*%s*." % (flow_name, prev_user)
        )
        obj.echo(
            "To delete this flow, you need to use the same production token that they used."
        )
        obj.echo(
            "Please reach out to them to get the token. Once you "
            "have it, call this command:"
        )
        obj.echo("    step-functions delete --authorize MY_TOKEN", fg="green")
        obj.echo(
            'See "Organizing Results" at docs.metaflow.org for more '
            "information about production tokens."
        )

    validate_token(
        obj.state_machine_name, obj.token_prefix, authorize, _token_instructions
    )

    obj.echo(
        "Deleting AWS Step Functions state machine *{name}*...".format(
            name=obj.state_machine_name
        ),
        bold=True,
    )
    schedule_deleted, sfn_deleted = StepFunctions.delete(obj.state_machine_name)

    if schedule_deleted:
        obj.echo(
            "Deleting Amazon EventBridge rule *{name}* as well...".format(
                name=obj.state_machine_name
            ),
            bold=True,
        )
    if sfn_deleted:
        obj.echo(
            "Deleting the AWS Step Functions state machine may take a while. "
            "Deploying the flow again to AWS Step Functions while the delete is in-flight will fail."
        )
        obj.echo(
            "In-flight executions will not be affected. "
            "If necessary, terminate them manually."
        )


@step_functions.command(help="Terminate flow execution on Step Functions.")
@click.option(
    "--authorize",
    default=None,
    type=str,
    help="Authorize the termination with a production token",
)
@click.argument("run-id", required=True, type=str)
@click.pass_obj
def terminate(obj, run_id, authorize=None):
    def _token_instructions(flow_name, prev_user):
        obj.echo(
            "There is an existing version of *%s* on AWS Step Functions which was "
            "deployed by the user *%s*." % (flow_name, prev_user)
        )
        obj.echo(
            "To terminate this flow, you need to use the same production token that they used."
        )
        obj.echo(
            "Please reach out to them to get the token. Once you have it, call "
            "this command:"
        )
        obj.echo("    step-functions terminate --authorize MY_TOKEN RUN_ID", fg="green")
        obj.echo(
            'See "Organizing Results" at docs.metaflow.org for more information '
            "about production tokens."
        )

    validate_run_id(
        obj.state_machine_name, obj.token_prefix, authorize, run_id, _token_instructions
    )

    # Trim prefix from run_id
    name = run_id[4:]
    obj.echo(
        "Terminating run *{run_id}* for {flow_name} ...".format(
            run_id=run_id, flow_name=obj.flow.name
        ),
        bold=True,
    )

    terminated = StepFunctions.terminate(obj.state_machine_name, name)
    if terminated:
        obj.echo("\nRun terminated at %s." % terminated.get("stopDate"))


def validate_run_id(
    state_machine_name, token_prefix, authorize, run_id, instructions_fn=None
):
    if not run_id.startswith("sfn-"):
        raise RunIdMismatch(
            "Run IDs for flows executed through AWS Step Functions begin with 'sfn-'"
        )

    name = run_id[4:]
    execution = StepFunctions.get_execution(state_machine_name, name)
    if execution is None:
        raise MetaflowException(
            "Could not find the execution *%s* (in RUNNING state) for the state machine *%s* on AWS Step Functions"
            % (name, state_machine_name)
        )

    _, owner, token, _ = execution

    if authorize is None:
        authorize = load_token(token_prefix)
    elif authorize.startswith("production:"):
        authorize = authorize[11:]

    if owner != get_username() and authorize != token:
        if instructions_fn:
            instructions_fn(flow_name=name, prev_user=owner)
        raise IncorrectProductionToken("Try again with the correct production token.")

    return True


def validate_token(name, token_prefix, authorize, instruction_fn=None):
    """
    Validate that the production token matches that of the deployed flow.

    In case both the user and token do not match, raises an error.
    Optionally outputs instructions on token usage via the provided instruction_fn(flow_name, prev_user)
    """
    # TODO: Unify this with the existing resolve_token implementation.

    # 1) retrieve the previous deployment, if one exists
    workflow = StepFunctions.get_existing_deployment(name)
    if workflow is None:
        prev_token = None
    else:
        prev_user, prev_token = workflow

    # 2) authorize this deployment
    if prev_token is not None:
        if authorize is None:
            authorize = load_token(token_prefix)
        elif authorize.startswith("production:"):
            authorize = authorize[11:]

        # we allow the user who deployed the previous version to re-deploy,
        # even if they don't have the token
        # NOTE: The username is visible in multiple sources, and can be set by the user.
        # Should we consider being stricter here?
        if prev_user != get_username() and authorize != prev_token:
            if instruction_fn:
                instruction_fn(flow_name=name, prev_user=prev_user)
            raise IncorrectProductionToken(
                "Try again with the correct production token."
            )

    # 3) all validations passed, store the previous token for future use
    token = prev_token

    store_token(token_prefix, token)
    return True



import json
import os
import shutil
import uuid
from concurrent.futures import as_completed
from tempfile import mkdtemp


from metaflow.datastore.datastore_storage import DataStoreStorage, CloseAfterUse
from metaflow.exception import MetaflowInternalError
from metaflow.metaflow_config import (
    DATASTORE_SYSROOT_GS,
    ARTIFACT_LOCALROOT,
    GS_STORAGE_WORKLOAD_TYPE,
)

from metaflow.plugins.gcp.gs_storage_client_factory import get_gs_storage_client
from metaflow.plugins.gcp.gs_utils import (
    check_gs_deps,
    parse_gs_full_path,
    process_gs_exception,
)
from metaflow.plugins.storage_executor import (
    StorageExecutor,
    handle_executor_exceptions,
)


class _GSRootClient(object):
    """
    datastore_root aware Google Cloud Storage client. I.e. blob operations are
    all done relative to datastore_root.

    This must be picklable, so methods may be passed across process boundaries.
    """

    def __init__(self, datastore_root):
        if datastore_root is None:
            raise MetaflowInternalError("datastore_root must be set")
        self._datastore_root = datastore_root

    def get_datastore_root(self):
        return self._datastore_root

    def get_blob_full_path(self, path):
        """
        Full path means <blob_prefix>/<path> where:
        datastore_root is gs://<bucket_name>/<blob_prefix>
        """
        _, blob_prefix = parse_gs_full_path(self._datastore_root)
        if blob_prefix is None:
            return path
        path = path.lstrip("/")
        return "/".join([blob_prefix, path])

    def get_bucket_client(self):
        bucket_name, _ = parse_gs_full_path(self._datastore_root)
        client = get_gs_storage_client()
        return client.bucket(bucket_name)

    def get_blob_client(self, path):
        bucket = self.get_bucket_client()
        blob_full_path = self.get_blob_full_path(path)
        blob = bucket.blob(blob_full_path)
        return blob

    # GS blob operations. These are meant to be single units of work
    # to be performed by thread or process pool workers.
    def is_file_single(self, path):
        """Drives GSStorage.is_file()"""
        try:
            blob = self.get_blob_client(path)
            result = blob.exists()

            return result
        except Exception as e:
            process_gs_exception(e)

    def list_content_single(self, path):
        """Drives GSStorage.list_content()"""

        def _trim_result(name, prefix):
            # Remove a prefix from the name, if present
            if prefix is not None and name[: len(prefix)] == prefix:
                name = name[len(prefix) + 1 :]
            return name

        try:
            path = path.rstrip("/") + "/"
            bucket_name, blob_prefix = parse_gs_full_path(self._datastore_root)
            full_path = self.get_blob_full_path(path)
            blobs = get_gs_storage_client().list_blobs(
                bucket_name,
                prefix=full_path,
                delimiter="/",
                include_trailing_delimiter=False,
            )
            result = []
            for b in blobs:
                result.append((_trim_result(b.name, blob_prefix), True))
            for p in blobs.prefixes:
                result.append((_trim_result(p, blob_prefix).rstrip("/"), False))
            return result
        except Exception as e:
            process_gs_exception(e)

    def save_bytes_single(
        self,
        path_tmpfile_metadata_triple,
        overwrite=False,
    ):
        try:
            path, tmpfile, metadata = path_tmpfile_metadata_triple
            blob = self.get_blob_client(path)
            if not overwrite:
                if blob.exists():
                    return
            if metadata is not None:
                blob.metadata = {"metaflow-user-attributes": json.dumps(metadata)}
            from google.cloud.storage.retry import DEFAULT_RETRY

            blob.upload_from_filename(
                tmpfile, retry=DEFAULT_RETRY, timeout=(14400, 60)
            )  # generous timeout for massive uploads. Use the same values as for Azure (connection_timeout, read_timeout)
        except Exception as e:
            process_gs_exception(e)

    def load_bytes_single(self, tmpdir, key):
        """Drives GSStorage.load_bytes()"""
        tmp_filename = os.path.join(tmpdir, str(uuid.uuid4()))
        blob = self.get_blob_client(key)
        metaflow_user_attributes = None
        import google.api_core.exceptions

        try:
            blob.reload()
            if blob.metadata and "metaflow-user-attributes" in blob.metadata:
                metaflow_user_attributes = json.loads(
                    blob.metadata["metaflow-user-attributes"]
                )
            blob.download_to_filename(tmp_filename)
        except google.api_core.exceptions.NotFound:
            tmp_filename = None
        return key, tmp_filename, metaflow_user_attributes


class GSStorage(DataStoreStorage):
    TYPE = "gs"

    @check_gs_deps
    def __init__(self, root=None):
        super(GSStorage, self).__init__(root)
        self._tmproot = ARTIFACT_LOCALROOT
        self._root_client = None

        self._use_processes = GS_STORAGE_WORKLOAD_TYPE == "high_throughput"
        self._executor = StorageExecutor(use_processes=self._use_processes)
        self._executor.warm_up()

    @property
    def root_client(self):
        """Root client is datastore_root aware. All blob operations go through root_client's
        methods.

        Method calls may run on the main thread, or be submitted to a process or thread pool.
        """
        if self._root_client is None:
            self._root_client = _GSRootClient(
                datastore_root=self.datastore_root,
            )
        return self._root_client

    @classmethod
    def get_datastore_root_from_config(cls, echo, create_on_absent=True):
        # create_on_absent doesn't do anything.  This matches S3Storage
        return DATASTORE_SYSROOT_GS

    @handle_executor_exceptions
    def is_file(self, paths):
        # preserving order is important...
        futures = [
            self._executor.submit(
                self.root_client.is_file_single,
                path,
            )
            for path in paths
        ]
        # preserving order is important...
        return [future.result() for future in futures]

    def info_file(self, path):
        # not used anywhere... we should consider killing this on all data storage implementations
        raise NotImplementedError()

    def size_file(self, path):
        import google.api_core.exceptions

        try:
            blob = self.root_client.get_blob_client(path)
            blob.reload()
            return blob.size
        except google.api_core.exceptions.NotFound:
            return None

    @handle_executor_exceptions
    def list_content(self, paths):
        futures = [
            self._executor.submit(self.root_client.list_content_single, path)
            for path in paths
        ]
        result = []
        for future in as_completed(futures):
            result.extend(self.list_content_result(*x) for x in future.result())
        return result

    @handle_executor_exceptions
    def save_bytes(self, path_and_bytes_iter, overwrite=False, len_hint=0):
        tmpdir = None
        try:
            tmpdir = mkdtemp(dir=ARTIFACT_LOCALROOT, prefix="metaflow.gs.save_bytes.")
            futures = []
            for path, byte_stream in path_and_bytes_iter:
                metadata = None
                # bytes_stream could actually be (bytes_stream, metadata) instead.
                # Read the small print on DatastoreStorage.save_bytes()
                if isinstance(byte_stream, tuple):
                    byte_stream, metadata = byte_stream
                tmp_filename = os.path.join(tmpdir, str(uuid.uuid4()))
                with open(tmp_filename, "wb") as f:
                    # make sure to close the file handle after reading.
                    with byte_stream as bytes:
                        f.write(bytes.read())
                # Fully finish writing the file, before submitting work. Careful with indentation.

                futures.append(
                    self._executor.submit(
                        self.root_client.save_bytes_single,
                        (path, tmp_filename, metadata),
                        overwrite=overwrite,
                    )
                )
            for future in as_completed(futures):
                future.result()
        finally:
            # *Future* improvement: We could clean up individual tmp files as each future completes
            if tmpdir and os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)

    @handle_executor_exceptions
    def load_bytes(self, keys):
        tmpdir = mkdtemp(dir=self._tmproot, prefix="metaflow.gs.load_bytes.")
        try:
            futures = [
                self._executor.submit(
                    self.root_client.load_bytes_single,
                    tmpdir,
                    key,
                )
                for key in keys
            ]

            # Let's detect any failures fast. Stop and cleanup ASAP.
            # Note that messing up return order is beneficial re: Hyrum's law too.
            items = [future.result() for future in as_completed(futures)]
        except Exception:
            # Other BaseExceptions will skip clean up here.
            # We let this go - smooth exit in those circumstances is more important than cleaning up a few tmp files
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            raise

        class _Closer(object):
            @staticmethod
            def close():
                if os.path.isdir(tmpdir):
                    shutil.rmtree(tmpdir)

        return CloseAfterUse(iter(items), closer=_Closer)



import json
import os
from datetime import datetime, timezone

###
# Algorithm to determine 1st error:
#   ignore the failures where message = ""
#   group the failures via templateName
#     sort each group by finishedAt
#   find the group for which the last finishedAt is earliest
#   if the earliest message is "No more retries left" then
#     get the n-1th message from that group
#   else
#     return the last message.
###


def parse_workflow_failures():
    failures = json.loads(
        json.loads(os.getenv("METAFLOW_ARGO_WORKFLOW_FAILURES", "[]"), strict=False),
        strict=False,
    )
    return [wf for wf in failures if wf.get("message")]


def group_failures_by_template(failures):
    groups = {}
    for failure in failures:
        if failure.get("finishedAt", None) is None:
            timestamp = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
            failure["finishedAt"] = timestamp
        groups.setdefault(failure["templateName"], []).append(failure)
    return groups


def sort_by_finished_at(items):
    return sorted(
        items, key=lambda x: datetime.strptime(x["finishedAt"], "%Y-%m-%dT%H:%M:%SZ")
    )


def find_earliest_last_finished_group(groups):
    return min(
        groups,
        key=lambda k: datetime.strptime(
            groups[k][-1]["finishedAt"], "%Y-%m-%dT%H:%M:%SZ"
        ),
    )


def determine_first_error():
    failures = parse_workflow_failures()
    if not failures:
        return None

    grouped_failures = group_failures_by_template(failures)
    for group in grouped_failures.values():
        group.sort(
            key=lambda g: datetime.strptime(g["finishedAt"], "%Y-%m-%dT%H:%M:%SZ")
        )

    earliest_group = grouped_failures[
        find_earliest_last_finished_group(grouped_failures)
    ]

    if earliest_group[-1]["message"] == "No more retries left":
        return earliest_group[-2]
    return earliest_group[-1]


if __name__ == "__main__":
    first_err = determine_first_error()
    print(json.dumps(first_err, indent=2))



from metaflow.exception import MetaflowException
from metaflow.decorators import FlowDecorator
from metaflow import current
from metaflow.util import get_username

import os
import re

# be careful when changing these limits. Other systems that see
# these names may rely on these limits
VALID_NAME_RE = "[^a-z0-9_]"
VALID_NAME_LEN = 128


class ProjectDecorator(FlowDecorator):
    """
    Specifies what flows belong to the same project.

    A project-specific namespace is created for all flows that
    use the same `@project(name)`.

    Parameters
    ----------
    name : str
        Project name. Make sure that the name is unique amongst all
        projects that use the same production scheduler. The name may
        contain only lowercase alphanumeric characters and underscores.

    branch : Optional[str], default None
        The branch to use. If not specified, the branch is set to
        `user.<username>` unless `production` is set to `True`. This can
        also be set on the command line using `--branch` as a top-level option.
        It is an error to specify `branch` in the decorator and on the command line.

    production : bool, default False
        Whether or not the branch is the production branch. This can also be set on the
        command line using `--production` as a top-level option. It is an error to specify
        `production` in the decorator and on the command line.
        The project branch name will be:
          - if `branch` is specified:
            - if `production` is True: `prod.<branch>`
            - if `production` is False: `test.<branch>`
          - if `branch` is not specified:
            - if `production` is True: `prod`
            - if `production` is False: `user.<username>`

    MF Add To Current
    -----------------
    project_name -> str
        The name of the project assigned to this flow, i.e. `X` in `@project(name=X)`.

        @@ Returns
        -------
        str
            Project name.

    project_flow_name -> str
        The flow name prefixed with the current project and branch. This name identifies
        the deployment on a production scheduler.

        @@ Returns
        -------
        str
            Flow name prefixed with project information.

    branch_name -> str
        The current branch, i.e. `X` in `--branch=X` set during deployment or run.

        @@ Returns
        -------
        str
            Branch name.

    is_user_branch -> bool
        True if the flow is deployed without a specific `--branch` or a `--production`
        flag.

        @@ Returns
        -------
        bool
            True if the deployment does not correspond to a specific branch.

    is_production -> bool
        True if the flow is deployed with the `--production` flag

        @@ Returns
        -------
        bool
            True if the flow is deployed with `--production`.
    """

    name = "project"

    options = {
        "production": dict(
            is_flag=True,
            default=False,
            show_default=True,
            help="Use the @project's production branch. To "
            "use a custom branch, use --branch.",
        ),
        "branch": dict(
            default=None,
            show_default=False,
            help="Use the given branch name under @project. "
            "The default is the user name if --production is "
            "not specified.",
        ),
    }

    defaults = {"name": None, **{k: v["default"] for k, v in options.items()}}

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        self._option_values = options
        project_name = self.attributes.get("name")
        for op in options:
            if (
                op in self._user_defined_attributes
                and options[op] != self.defaults[op]
                and self.attributes[op] != options[op]
            ):
                # Exception if:
                #  - the user provides a value in the attributes field
                #  - AND the user provided a value in the command line (non default)
                #  - AND the values are different
                # Note that this won't raise an error if the user provided the default
                # value in the command line and provided one in attribute but although
                # slightly inconsistent, it is not incorrect.
                raise MetaflowException(
                    "You cannot pass %s as both a command-line argument and an attribute "
                    "of the @project decorator." % op
                )
        if "branch" in self._user_defined_attributes:
            project_branch = self.attributes["branch"]
        else:
            project_branch = options["branch"]

        if "production" in self._user_defined_attributes:
            project_production = self.attributes["production"]
        else:
            project_production = options["production"]

        project_flow_name, branch_name = format_name(
            flow.name,
            project_name,
            project_production,
            project_branch,
            get_username(),
        )
        is_user_branch = project_branch is None and not project_production
        echo(
            "Project: *%s*, Branch: *%s*" % (project_name, branch_name),
            fg="magenta",
            highlight="green",
        )
        current._update_env(
            {
                "project_name": project_name,
                "branch_name": branch_name,
                "is_user_branch": is_user_branch,
                "is_production": project_production,
                "project_flow_name": project_flow_name,
            }
        )
        metadata.add_sticky_tags(
            sys_tags=["project:%s" % project_name, "project_branch:%s" % branch_name]
        )

    def get_top_level_options(self):
        return list(self._option_values.items())


def format_name(flow_name, project_name, deploy_prod, given_branch, user_name):
    if not project_name:
        # an empty string is not a valid project name
        raise MetaflowException(
            "@project needs a name. " "Try @project(name='some_name')"
        )
    elif re.search(VALID_NAME_RE, project_name):
        raise MetaflowException(
            "The @project name must contain only "
            "lowercase alphanumeric characters "
            "and underscores."
        )
    elif len(project_name) > VALID_NAME_LEN:
        raise MetaflowException(
            "The @project name must be shorter than " "%d characters." % VALID_NAME_LEN
        )

    if given_branch:
        if re.search(VALID_NAME_RE, given_branch):
            raise MetaflowException(
                "The branch name must contain only "
                "lowercase alphanumeric characters "
                "and underscores."
            )
        elif len(given_branch) > VALID_NAME_LEN:
            raise MetaflowException(
                "Branch name is too long. "
                "The maximum is %d characters." % VALID_NAME_LEN
            )
        if deploy_prod:
            branch = "prod.%s" % given_branch
        else:
            branch = "test.%s" % given_branch
    elif deploy_prod:
        branch = "prod"
    else:
        # For AWS Step Functions, we set the branch to the value of
        # environment variable `METAFLOW_OWNER`, since AWS Step Functions
        # has no notion of user name.
        branch = "user.%s" % os.environ.get("METAFLOW_OWNER", user_name)

    return ".".join((project_name, branch, flow_name)), branch



from __future__ import print_function
from collections import OrderedDict
import json
import os
import sys
import time
from tempfile import NamedTemporaryFile
from hashlib import sha1

from urllib.parse import urlparse

from metaflow.datastore import FlowDataStore
from metaflow.datastore.content_addressed_store import BlobCache
from metaflow.datastore.flow_datastore import MetadataCache
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    CLIENT_CACHE_PATH,
    CLIENT_CACHE_MAX_SIZE,
    CLIENT_CACHE_MAX_FLOWDATASTORE_COUNT,
)
from metaflow.metaflow_profile import from_start

from metaflow.plugins import DATASTORES

NEW_FILE_QUARANTINE = 10

if sys.version_info[0] >= 3 and sys.version_info[1] >= 2:

    def od_move_to_end(od, key):
        od.move_to_end(key)

else:
    # Not very efficient but works and most people are on 3.2+
    def od_move_to_end(od, key):
        v = od.get(key)
        del od[key]
        od[key] = v


class FileCacheException(MetaflowException):
    headline = "File cache error"


class FileCache(object):
    def __init__(self, cache_dir=None, max_size=None):
        self._cache_dir = cache_dir
        self._max_size = max_size
        if self._cache_dir is None:
            self._cache_dir = CLIENT_CACHE_PATH
        if self._max_size is None:
            self._max_size = int(CLIENT_CACHE_MAX_SIZE)
        self._total = 0

        self._objects = None
        # We have a separate blob_cache per flow and datastore type.
        self._blob_caches = {}

        # We also keep a cache for FlowDataStore objects because some of them
        # may have long-lived persistent connections; this is purely a
        # performance optimization. Uses OrderedDict to implement a kind of LRU
        # cache and keep only a certain number of these caches around.
        self._store_caches = OrderedDict()

        # We also keep a cache of data_metadata for TaskDatastore. This is used
        # when querying for sizes of artifacts. Once we have queried for the size
        # of one artifact in a TaskDatastore, caching this means that any
        # queries on that same TaskDatastore will be quick (since we already
        # have all the metadata). We keep track of this in a file so it persists
        # across processes.

    @property
    def cache_dir(self):
        return self._cache_dir

    def get_logs_stream(
        self, ds_type, ds_root, stream, attempt, flow_name, run_id, step_name, task_id
    ):
        from metaflow.mflog import LOG_SOURCES

        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        task_ds = ds.get_task_datastore(
            run_id, step_name, task_id, data_metadata={"objects": {}, "info": {}}
        )
        return task_ds.load_logs(LOG_SOURCES, stream, attempt_override=attempt)

    def get_log_legacy(
        self, ds_type, location, logtype, attempt, flow_name, run_id, step_name, task_id
    ):
        ds_cls = self._get_datastore_storage_impl(ds_type)
        ds_root = ds_cls.path_join(*ds_cls.path_split(location)[:-5])
        cache_id = self.flow_ds_id(ds_type, ds_root, flow_name)

        token = (
            "%s.cached"
            % sha1(
                os.path.join(run_id, step_name, task_id, "%s_log" % logtype).encode(
                    "utf-8"
                )
            ).hexdigest()
        )
        path = os.path.join(self._cache_dir, cache_id, token[:2], token)

        cached_log = self.read_file(path)
        if cached_log is not None:
            return cached_log

        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        task_ds = ds.get_task_datastore(
            run_id, step_name, task_id, data_metadata={"objects": {}, "info": {}}
        )

        log = task_ds.load_log_legacy(logtype, attempt_override=attempt)
        # Store this in the file cache as well
        self.create_file(path, log)
        return log

    def get_legacy_log_size(
        self, ds_type, location, logtype, attempt, flow_name, run_id, step_name, task_id
    ):
        ds_cls = self._get_datastore_storage_impl(ds_type)
        ds_root = ds_cls.path_join(*ds_cls.path_split(location)[:-5])
        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        task_ds = ds.get_task_datastore(
            run_id,
            step_name,
            task_id,
            attempt=attempt,
            data_metadata={"objects": {}, "info": {}},
        )

        return task_ds.get_legacy_log_size(logtype)

    def get_log_size(
        self, ds_type, ds_root, logtype, attempt, flow_name, run_id, step_name, task_id
    ):
        from metaflow.mflog import LOG_SOURCES

        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        task_ds = ds.get_task_datastore(
            run_id,
            step_name,
            task_id,
            attempt=attempt,
            data_metadata={"objects": {}, "info": {}},
        )

        return task_ds.get_log_size(LOG_SOURCES, logtype)

    def get_data(self, ds_type, flow_name, location, key):
        ds_cls = self._get_datastore_storage_impl(ds_type)
        ds_root = ds_cls.get_datastore_root_from_location(location, flow_name)
        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        return next(ds.load_data([key], force_raw=True))

    def get_artifact_size_by_location(
        self, ds_type, location, attempt, flow_name, run_id, step_name, task_id, name
    ):
        """Gets the size of the artifact content (in bytes) for the name at the location"""
        ds_cls = self._get_datastore_storage_impl(ds_type)
        ds_root = ds_cls.get_datastore_root_from_location(location, flow_name)

        return self.get_artifact_size(
            ds_type, ds_root, attempt, flow_name, run_id, step_name, task_id, name
        )

    def get_artifact_size(
        self, ds_type, ds_root, attempt, flow_name, run_id, step_name, task_id, name
    ):
        """Gets the size of the artifact content (in bytes) for the name"""
        task_ds = self._get_task_datastore(
            ds_type, ds_root, flow_name, run_id, step_name, task_id, attempt
        )

        _, size = next(task_ds.get_artifact_sizes([name]))
        return size

    def get_artifact_by_location(
        self,
        ds_type,
        location,
        data_metadata,
        flow_name,
        run_id,
        step_name,
        task_id,
        name,
    ):
        ds_cls = self._get_datastore_storage_impl(ds_type)
        ds_root = ds_cls.get_datastore_root_from_location(location, flow_name)
        return self.get_artifact(
            ds_type, ds_root, data_metadata, flow_name, run_id, step_name, task_id, name
        )

    def get_artifact(
        self,
        ds_type,
        ds_root,
        data_metadata,
        flow_name,
        run_id,
        step_name,
        task_id,
        name,
    ):
        _, obj = next(
            self.get_artifacts(
                ds_type,
                ds_root,
                data_metadata,
                flow_name,
                run_id,
                step_name,
                task_id,
                [name],
            )
        )
        return obj

    def get_all_artifacts(
        self, ds_type, ds_root, data_metadata, flow_name, run_id, step_name, task_id
    ):
        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        # We get the task datastore for this task
        task_ds = ds.get_task_datastore(
            run_id, step_name, task_id, data_metadata=data_metadata
        )
        # This will reuse the blob cache if needed. We do not have an
        # artifact cache so the unpickling happens every time here.
        return task_ds.load_artifacts([n for n, _ in task_ds.items()])

    def get_artifacts(
        self,
        ds_type,
        ds_root,
        data_metadata,
        flow_name,
        run_id,
        step_name,
        task_id,
        names,
    ):
        ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        # We get the task datastore for this task
        task_ds = ds.get_task_datastore(
            run_id, step_name, task_id, data_metadata=data_metadata
        )
        # note that load_artifacts uses flow_datastore.castore which goes
        # through one of the self._blob_cache
        return task_ds.load_artifacts(names)

    def create_file(self, path, value):
        if self._objects is None:
            # Index objects lazily (when we first need to write to it).
            # This can be an expensive operation
            self._index_objects()
        dirname = os.path.dirname(path)
        try:
            FileCache._makedirs(dirname)
        except:  # noqa E722
            raise FileCacheException("Could not create directory: %s" % dirname)
        tmpfile = NamedTemporaryFile(dir=dirname, prefix="dlobj", delete=False)
        # Now write out the file
        try:
            tmpfile.write(value)
            tmpfile.flush()
            os.rename(tmpfile.name, path)
        except:  # noqa E722
            os.unlink(tmpfile.name)
            raise
        size = os.path.getsize(path)
        self._total += size
        self._objects.append((int(time.time()), size, path))
        self._garbage_collect()

    def read_file(self, path):
        if os.path.exists(path):
            try:
                with open(path, "rb") as f:
                    return f.read()
            except IOError:
                # It may have been concurrently garbage collected by another
                # process
                pass
        return None

    def _index_objects(self):
        objects = []
        if os.path.exists(self._cache_dir):
            for flow_ds_id in os.listdir(self._cache_dir):
                root = os.path.join(self._cache_dir, flow_ds_id)
                if not os.path.isdir(root):
                    continue
                for subdir in os.listdir(root):
                    root = os.path.join(self._cache_dir, flow_ds_id, subdir)
                    if not os.path.isdir(root):
                        continue
                    for obj in os.listdir(root):
                        sha, ext = os.path.splitext(obj)
                        if ext in ["cached", "blob"]:
                            path = os.path.join(root, obj)
                            objects.insert(
                                0, (os.path.getctime(path), os.path.getsize(path), path)
                            )

        self._total = sum(size for _, size, _ in objects)
        self._objects = sorted(objects, reverse=False)

    @staticmethod
    def flow_ds_id(ds_type, ds_root, flow_name):
        p = urlparse(ds_root)
        sanitized_root = (p.netloc + p.path).replace("/", "_")
        return ".".join([ds_type, sanitized_root, flow_name])

    @staticmethod
    def task_ds_id(ds_type, ds_root, flow_name, run_id, step_name, task_id, attempt):
        p = urlparse(ds_root)
        sanitized_root = (p.netloc + p.path).replace("/", "_")
        return ".".join(
            [
                ds_type,
                sanitized_root,
                flow_name,
                run_id,
                step_name,
                task_id,
                str(attempt),
            ]
        )

    def _garbage_collect(self):
        now = time.time()
        while self._objects and self._total > self._max_size * 1024**2:
            if now - self._objects[0][0] < NEW_FILE_QUARANTINE:
                break
            ctime, size, path = self._objects.pop(0)
            self._total -= size
            try:
                os.remove(path)
            except OSError:
                # maybe another client had already GC'ed the file away
                pass

    @staticmethod
    def _makedirs(path):
        # this is for python2 compatibility.
        # Python3 has os.makedirs(exist_ok=True).
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                return
            else:
                raise

    @staticmethod
    def _get_datastore_storage_impl(ds_type):
        storage_impl = [d for d in DATASTORES if d.TYPE == ds_type]
        if len(storage_impl) == 0:
            raise FileCacheException("Datastore %s was not found" % ds_type)
        return storage_impl[0]

    def _get_flow_datastore(self, ds_type, ds_root, flow_name):
        cache_id = self.flow_ds_id(ds_type, ds_root, flow_name)
        cached_flow_datastore = self._store_caches.get(cache_id)

        if cached_flow_datastore:
            od_move_to_end(self._store_caches, cache_id)
            return cached_flow_datastore
        else:
            storage_impl = self._get_datastore_storage_impl(ds_type)
            cached_flow_datastore = FlowDataStore(
                flow_name=flow_name,
                environment=None,  # TODO: Add environment here
                storage_impl=storage_impl,
                ds_root=ds_root,
            )
            blob_cache = self._blob_caches.setdefault(
                cache_id,
                (
                    FileBlobCache(self, cache_id),
                    TaskMetadataCache(self, ds_type, ds_root, flow_name),
                ),
            )
            cached_flow_datastore.ca_store.set_blob_cache(blob_cache[0])
            cached_flow_datastore.set_metadata_cache(blob_cache[1])
            self._store_caches[cache_id] = cached_flow_datastore
            if len(self._store_caches) > CLIENT_CACHE_MAX_FLOWDATASTORE_COUNT:
                cache_id_to_remove, _ = self._store_caches.popitem(last=False)
                del self._blob_caches[cache_id_to_remove]
            return cached_flow_datastore

    def _get_task_datastore(
        self, ds_type, ds_root, flow_name, run_id, step_name, task_id, attempt
    ):
        flow_ds = self._get_flow_datastore(ds_type, ds_root, flow_name)

        return flow_ds.get_task_datastore(run_id, step_name, task_id, attempt=attempt)


class TaskMetadataCache(MetadataCache):
    def __init__(self, filecache, ds_type, ds_root, flow_name):
        self._filecache = filecache
        self._ds_type = ds_type
        self._ds_root = ds_root
        self._flow_name = flow_name

    def _path(self, run_id, step_name, task_id, attempt):
        if attempt is None:
            raise MetaflowException(
                "Attempt number must be specified to use task metadata cache. Raise an issue "
                "on Metaflow GitHub if you see this message.",
            )
        cache_id = self._filecache.task_ds_id(
            self._ds_type,
            self._ds_root,
            self._flow_name,
            run_id,
            step_name,
            task_id,
            attempt,
        )
        token = (
            "%s.cached"
            % sha1(
                os.path.join(
                    run_id, step_name, task_id, str(attempt), "metadata"
                ).encode("utf-8")
            ).hexdigest()
        )
        return os.path.join(self._filecache.cache_dir, cache_id, token[:2], token)

    def load_metadata(self, run_id, step_name, task_id, attempt):
        d = self._filecache.read_file(self._path(run_id, step_name, task_id, attempt))
        if d:
            return json.loads(d)

    def store_metadata(self, run_id, step_name, task_id, attempt, metadata_dict):
        self._filecache.create_file(
            self._path(run_id, step_name, task_id, attempt),
            json.dumps(metadata_dict).encode("utf-8"),
        )


class FileBlobCache(BlobCache):
    def __init__(self, filecache, cache_id):
        self._filecache = filecache
        self._cache_id = cache_id

    def _path(self, key):
        key_dir = key[:2]
        return os.path.join(
            self._filecache.cache_dir, self._cache_id, key_dir, "%s.blob" % key
        )

    def load_key(self, key):
        return self._filecache.read_file(self._path(key))

    def store_key(self, key, blob):
        self._filecache.create_file(self._path(key), blob)



import os

from itertools import starmap

from metaflow.plugins.datatools.s3.s3 import S3, S3Client, S3PutObject, check_s3_deps
from metaflow.metaflow_config import DATASTORE_SYSROOT_S3, ARTIFACT_LOCALROOT
from metaflow.datastore.datastore_storage import CloseAfterUse, DataStoreStorage


try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse


class S3Storage(DataStoreStorage):
    TYPE = "s3"

    @check_s3_deps
    def __init__(self, root=None):
        super(S3Storage, self).__init__(root)
        self.s3_client = S3Client()

    @classmethod
    def get_datastore_root_from_config(cls, echo, create_on_absent=True):
        return DATASTORE_SYSROOT_S3

    def is_file(self, paths):
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            external_client=self.s3_client,
        ) as s3:
            if len(paths) > 10:
                s3objs = s3.info_many(paths, return_missing=True)
                return [s3obj.exists for s3obj in s3objs]
            else:
                result = []
                for path in paths:
                    result.append(s3.info(path, return_missing=True).exists)
                return result

    def info_file(self, path):
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            external_client=self.s3_client,
        ) as s3:
            s3obj = s3.info(path, return_missing=True)
            return s3obj.exists, s3obj.metadata

    def size_file(self, path):
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            external_client=self.s3_client,
        ) as s3:
            s3obj = s3.info(path, return_missing=True)
            return s3obj.size

    def list_content(self, paths):
        strip_prefix_len = len(self.datastore_root.rstrip("/")) + 1
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            external_client=self.s3_client,
        ) as s3:
            results = s3.list_paths(paths)
            return [
                self.list_content_result(
                    path=o.url[strip_prefix_len:], is_file=o.exists
                )
                for o in results
            ]

    def save_bytes(self, path_and_bytes_iter, overwrite=False, len_hint=0):
        def _convert():
            # Output format is the same as what is needed for S3PutObject:
            # key, value, path, content_type, encryption, metadata
            for path, obj in path_and_bytes_iter:
                if isinstance(obj, tuple):
                    yield path, obj[0], None, None, None, obj[1]
                else:
                    yield path, obj, None, None, None, None

        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            external_client=self.s3_client,
        ) as s3:
            # HACK: The S3 datatools we rely on does not currently do a good job
            # determining if uploading things in parallel is more efficient than
            # serially. We use a heuristic for now where if we have a lot of
            # files, we will go in parallel and if we have few files, we will
            # serially upload them. This is not ideal because there is also a size
            # factor and one very large file with a few other small files, for
            # example, would benefit from a parallel upload.
            #
            # In the case of save_artifacts, currently len_hint is based on the
            # total number of artifacts, not taking into account how many of them
            # already exist in the CAS, i.e. it can be a gross overestimate. As a
            # result, it is possible we take a latency hit by using put_many only
            # for a few artifacts.
            #
            # A better approach would be to e.g. write all blobs to temp files
            # and based on the total size and number of files use either put_files
            # (which avoids re-writing the files) or s3.put sequentially.
            if len_hint > 10:
                # Use put_many
                s3.put_many(starmap(S3PutObject, _convert()), overwrite)
            else:
                # Sequential upload
                for key, obj, _, _, _, metadata in _convert():
                    s3.put(key, obj, overwrite=overwrite, metadata=metadata)

    def load_bytes(self, paths):
        if len(paths) == 0:
            return CloseAfterUse(iter([]))

        s3 = S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            external_client=self.s3_client,
        )

        def iter_results():
            # We similarly do things in parallel for many files. This is again
            # a hack.
            if len(paths) > 10:
                results = s3.get_many(paths, return_missing=True, return_info=True)
                for r in results:
                    if r.exists:
                        yield r.key, r.path, r.metadata
                    else:
                        yield r.key, None, None
            else:
                for p in paths:
                    r = s3.get(p, return_missing=True, return_info=True)
                    if r.exists:
                        yield r.key, r.path, r.metadata
                    else:
                        yield r.key, None, None

        return CloseAfterUse(iter_results(), closer=s3)



import os
import json

# Because Pip does not offer a direct way to set target platform_system and platform_machine values for resolving packages transitive dependencies, we need to instead
# manually patch the correct target architecture values for pip to be able to resolve the whole dependency tree successfully.
# This is necessary for packages that have conditional dependencies dependent on machine/system, e.g. Torch

try:
    import pip._vendor.packaging.markers as _markers
except ImportError:
    _markers = None

_overrides = os.environ.get("PIP_CUSTOMIZE_OVERRIDES")
if _overrides and _markers:
    _orig_default_environment = _markers.default_environment
    _overrides_dict = json.loads(_overrides)

    def _wrap_default_environment():
        result = _orig_default_environment()
        result.update(_overrides_dict)
        return result

    _markers.default_environment = _wrap_default_environment



from __future__ import annotations

import ast
import inspect
import sys
from collections.abc import Sequence
from functools import partial
from inspect import isclass, isfunction
from types import CodeType, FrameType, FunctionType
from typing import TYPE_CHECKING, Any, Callable, ForwardRef, TypeVar, cast, overload
from warnings import warn

from ._config import CollectionCheckStrategy, ForwardRefPolicy, global_config
from ._exceptions import InstrumentationWarning
from ._functions import TypeCheckFailCallback
from ._transformer import TypeguardTransformer
from ._utils import Unset, function_name, get_stacklevel, is_method_of, unset

T_CallableOrType = TypeVar("T_CallableOrType", bound=Callable[..., Any])

if TYPE_CHECKING:
    from typeshed.stdlib.types import _Cell

    def typeguard_ignore(f: T_CallableOrType) -> T_CallableOrType:
        """This decorator is a noop during static type-checking."""
        return f

else:
    from typing import no_type_check as typeguard_ignore  # noqa: F401


def make_cell(value: object) -> _Cell:
    return (lambda: value).__closure__[0]  # type: ignore[index]


def find_target_function(
    new_code: CodeType, target_path: Sequence[str], firstlineno: int
) -> CodeType | None:
    target_name = target_path[0]
    for const in new_code.co_consts:
        if isinstance(const, CodeType):
            if const.co_name == target_name:
                if const.co_firstlineno == firstlineno:
                    return const
                elif len(target_path) > 1:
                    target_code = find_target_function(
                        const, target_path[1:], firstlineno
                    )
                    if target_code:
                        return target_code

    return None


def instrument(f: T_CallableOrType) -> FunctionType | str:
    if not getattr(f, "__code__", None):
        return "no code associated"
    elif not getattr(f, "__module__", None):
        return "__module__ attribute is not set"
    elif f.__code__.co_filename == "<stdin>":
        return "cannot instrument functions defined in a REPL"
    elif hasattr(f, "__wrapped__"):
        return (
            "@typechecked only supports instrumenting functions wrapped with "
            "@classmethod, @staticmethod or @property"
        )

    target_path = [item for item in f.__qualname__.split(".") if item != "<locals>"]
    module_source = inspect.getsource(sys.modules[f.__module__])
    module_ast = ast.parse(module_source)
    instrumentor = TypeguardTransformer(target_path, f.__code__.co_firstlineno)
    instrumentor.visit(module_ast)

    if not instrumentor.target_node or instrumentor.target_lineno is None:
        return "instrumentor did not find the target function"

    module_code = compile(module_ast, f.__code__.co_filename, "exec", dont_inherit=True)
    new_code = find_target_function(
        module_code, target_path, instrumentor.target_lineno
    )
    if not new_code:
        return "cannot find the target function in the AST"

    if global_config.debug_instrumentation and sys.version_info >= (3, 9):
        # Find the matching AST node, then unparse it to source and print to stdout
        print(
            f"Source code of {f.__qualname__}() after instrumentation:"
            "\n----------------------------------------------",
            file=sys.stderr,
        )
        print(ast.unparse(instrumentor.target_node), file=sys.stderr)
        print(
            "----------------------------------------------",
            file=sys.stderr,
        )

    closure = f.__closure__
    if new_code.co_freevars != f.__code__.co_freevars:
        # Create a new closure and find values for the new free variables
        frame = cast(FrameType, inspect.currentframe())
        frame = cast(FrameType, frame.f_back)
        frame_locals = cast(FrameType, frame.f_back).f_locals
        cells: list[_Cell] = []
        for key in new_code.co_freevars:
            if key in instrumentor.names_used_in_annotations:
                # Find the value and make a new cell from it
                value = frame_locals.get(key) or ForwardRef(key)
                cells.append(make_cell(value))
            else:
                # Reuse the cell from the existing closure
                assert f.__closure__
                cells.append(f.__closure__[f.__code__.co_freevars.index(key)])

        closure = tuple(cells)

    new_function = FunctionType(new_code, f.__globals__, f.__name__, closure=closure)
    new_function.__module__ = f.__module__
    new_function.__name__ = f.__name__
    new_function.__qualname__ = f.__qualname__
    new_function.__annotations__ = f.__annotations__
    new_function.__doc__ = f.__doc__
    new_function.__defaults__ = f.__defaults__
    new_function.__kwdefaults__ = f.__kwdefaults__
    return new_function


@overload
def typechecked(
    *,
    forward_ref_policy: ForwardRefPolicy | Unset = unset,
    typecheck_fail_callback: TypeCheckFailCallback | Unset = unset,
    collection_check_strategy: CollectionCheckStrategy | Unset = unset,
    debug_instrumentation: bool | Unset = unset,
) -> Callable[[T_CallableOrType], T_CallableOrType]: ...


@overload
def typechecked(target: T_CallableOrType) -> T_CallableOrType: ...


def typechecked(
    target: T_CallableOrType | None = None,
    *,
    forward_ref_policy: ForwardRefPolicy | Unset = unset,
    typecheck_fail_callback: TypeCheckFailCallback | Unset = unset,
    collection_check_strategy: CollectionCheckStrategy | Unset = unset,
    debug_instrumentation: bool | Unset = unset,
) -> Any:
    """
    Instrument the target function to perform run-time type checking.

    This decorator recompiles the target function, injecting code to type check
    arguments, return values, yield values (excluding ``yield from``) and assignments to
    annotated local variables.

    This can also be used as a class decorator. This will instrument all type annotated
    methods, including :func:`@classmethod <classmethod>`,
    :func:`@staticmethod <staticmethod>`,  and :class:`@property <property>` decorated
    methods in the class.

    .. note:: When Python is run in optimized mode (``-O`` or ``-OO``, this decorator
        is a no-op). This is a feature meant for selectively introducing type checking
        into a code base where the checks aren't meant to be run in production.

    :param target: the function or class to enable type checking for
    :param forward_ref_policy: override for
        :attr:`.TypeCheckConfiguration.forward_ref_policy`
    :param typecheck_fail_callback: override for
        :attr:`.TypeCheckConfiguration.typecheck_fail_callback`
    :param collection_check_strategy: override for
        :attr:`.TypeCheckConfiguration.collection_check_strategy`
    :param debug_instrumentation: override for
        :attr:`.TypeCheckConfiguration.debug_instrumentation`

    """
    if target is None:
        return partial(
            typechecked,
            forward_ref_policy=forward_ref_policy,
            typecheck_fail_callback=typecheck_fail_callback,
            collection_check_strategy=collection_check_strategy,
            debug_instrumentation=debug_instrumentation,
        )

    if not __debug__:
        return target

    if isclass(target):
        for key, attr in target.__dict__.items():
            if is_method_of(attr, target):
                retval = instrument(attr)
                if isfunction(retval):
                    setattr(target, key, retval)
            elif isinstance(attr, (classmethod, staticmethod)):
                if is_method_of(attr.__func__, target):
                    retval = instrument(attr.__func__)
                    if isfunction(retval):
                        wrapper = attr.__class__(retval)
                        setattr(target, key, wrapper)
            elif isinstance(attr, property):
                kwargs: dict[str, Any] = dict(doc=attr.__doc__)
                for name in ("fset", "fget", "fdel"):
                    property_func = kwargs[name] = getattr(attr, name)
                    if is_method_of(property_func, target):
                        retval = instrument(property_func)
                        if isfunction(retval):
                            kwargs[name] = retval

                setattr(target, key, attr.__class__(**kwargs))

        return target

    # Find either the first Python wrapper or the actual function
    wrapper_class: (
        type[classmethod[Any, Any, Any]] | type[staticmethod[Any, Any]] | None
    ) = None
    if isinstance(target, (classmethod, staticmethod)):
        wrapper_class = target.__class__
        target = target.__func__

    retval = instrument(target)
    if isinstance(retval, str):
        warn(
            f"{retval} -- not typechecking {function_name(target)}",
            InstrumentationWarning,
            stacklevel=get_stacklevel(),
        )
        return target

    if wrapper_class is None:
        return retval
    else:
        return wrapper_class(retval)



from __future__ import annotations

import inspect
import sys
from importlib import import_module
from inspect import currentframe
from types import CodeType, FrameType, FunctionType
from typing import TYPE_CHECKING, Any, Callable, ForwardRef, Union, cast, final
from weakref import WeakValueDictionary

if TYPE_CHECKING:
    from ._memo import TypeCheckMemo

if sys.version_info >= (3, 13):
    from typing import get_args, get_origin

    def evaluate_forwardref(forwardref: ForwardRef, memo: TypeCheckMemo) -> Any:
        return forwardref._evaluate(
            memo.globals, memo.locals, type_params=(), recursive_guard=frozenset()
        )

elif sys.version_info >= (3, 10):
    from typing import get_args, get_origin

    def evaluate_forwardref(forwardref: ForwardRef, memo: TypeCheckMemo) -> Any:
        return forwardref._evaluate(
            memo.globals, memo.locals, recursive_guard=frozenset()
        )

else:
    from metaflow._vendor.typing_extensions import get_args, get_origin

    evaluate_extra_args: tuple[frozenset[Any], ...] = (
        (frozenset(),) if sys.version_info >= (3, 9) else ()
    )

    def evaluate_forwardref(forwardref: ForwardRef, memo: TypeCheckMemo) -> Any:
        from ._union_transformer import compile_type_hint, type_substitutions

        if not forwardref.__forward_evaluated__:
            forwardref.__forward_code__ = compile_type_hint(forwardref.__forward_arg__)

        try:
            return forwardref._evaluate(memo.globals, memo.locals, *evaluate_extra_args)
        except NameError:
            if sys.version_info < (3, 10):
                # Try again, with the type substitutions (list -> List etc.) in place
                new_globals = memo.globals.copy()
                new_globals.setdefault("Union", Union)
                if sys.version_info < (3, 9):
                    new_globals.update(type_substitutions)

                return forwardref._evaluate(
                    new_globals, memo.locals or new_globals, *evaluate_extra_args
                )

            raise


_functions_map: WeakValueDictionary[CodeType, FunctionType] = WeakValueDictionary()


def get_type_name(type_: Any) -> str:
    name: str
    for attrname in "__name__", "_name", "__forward_arg__":
        candidate = getattr(type_, attrname, None)
        if isinstance(candidate, str):
            name = candidate
            break
    else:
        origin = get_origin(type_)
        candidate = getattr(origin, "_name", None)
        if candidate is None:
            candidate = type_.__class__.__name__.strip("_")

        if isinstance(candidate, str):
            name = candidate
        else:
            return "(unknown)"

    args = get_args(type_)
    if args:
        if name == "Literal":
            formatted_args = ", ".join(repr(arg) for arg in args)
        else:
            formatted_args = ", ".join(get_type_name(arg) for arg in args)

        name += f"[{formatted_args}]"

    module = getattr(type_, "__module__", None)
    if module and module not in (None, "typing", "typing_extensions", "builtins"):
        name = module + "." + name

    return name


def qualified_name(obj: Any, *, add_class_prefix: bool = False) -> str:
    """
    Return the qualified name (e.g. package.module.Type) for the given object.

    Builtins and types from the :mod:`typing` package get special treatment by having
    the module name stripped from the generated name.

    """
    if obj is None:
        return "None"
    elif inspect.isclass(obj):
        prefix = "class " if add_class_prefix else ""
        type_ = obj
    else:
        prefix = ""
        type_ = type(obj)

    module = type_.__module__
    qualname = type_.__qualname__
    name = qualname if module in ("typing", "builtins") else f"{module}.{qualname}"
    return prefix + name


def function_name(func: Callable[..., Any]) -> str:
    """
    Return the qualified name of the given function.

    Builtins and types from the :mod:`typing` package get special treatment by having
    the module name stripped from the generated name.

    """
    # For partial functions and objects with __call__ defined, __qualname__ does not
    # exist
    module = getattr(func, "__module__", "")
    qualname = (module + ".") if module not in ("builtins", "") else ""
    return qualname + getattr(func, "__qualname__", repr(func))


def resolve_reference(reference: str) -> Any:
    modulename, varname = reference.partition(":")[::2]
    if not modulename or not varname:
        raise ValueError(f"{reference!r} is not a module:varname reference")

    obj = import_module(modulename)
    for attr in varname.split("."):
        obj = getattr(obj, attr)

    return obj


def is_method_of(obj: object, cls: type) -> bool:
    return (
        inspect.isfunction(obj)
        and obj.__module__ == cls.__module__
        and obj.__qualname__.startswith(cls.__qualname__ + ".")
    )


def get_stacklevel() -> int:
    level = 1
    frame = cast(FrameType, currentframe()).f_back
    while frame and frame.f_globals.get("__name__", "").startswith("typeguard."):
        level += 1
        frame = frame.f_back

    return level


@final
class Unset:
    __slots__ = ()

    def __repr__(self) -> str:
        return "<unset>"


unset = Unset()



import importlib
import inspect
import os
import sys
import json

from typing import Dict, Iterator, Optional, Tuple

from metaflow import Run, Task

from metaflow.metaflow_config import CLICK_API_PROCESS_CONFIG

from metaflow.plugins import get_runner_cli

from .utils import (
    temporary_fifo,
    handle_timeout,
    async_handle_timeout,
    with_dir,
)
from .subprocess_manager import CommandManager, SubprocessManager


class ExecutingProcess(object):
    """
    This is a base class for `ExecutingRun` and `ExecutingTask` classes.
    The `ExecutingRun` and `ExecutingTask` classes are returned by methods
    in `Runner` and `NBRunner`, and they are subclasses of this class.

    The `ExecutingRun` class for instance contains a reference to a `metaflow.Run`
    object representing the currently executing or finished run, as well as the metadata
    related to the process.

    Similarly, the `ExecutingTask` class contains a reference to a `metaflow.Task`
    object representing the currently executing or finished task, as well as the metadata
    related to the process.

    This class or its subclasses are not meant to be instantiated directly. The class
    works as a context manager, allowing you to use a pattern like:

    ```python
    with Runner(...).run() as running:
        ...
    ```

    Note that you should use either this object as the context manager or `Runner`, not both
    in a nested manner.
    """

    def __init__(self, runner: "Runner", command_obj: CommandManager) -> None:
        """
        Create a new ExecutingRun -- this should not be done by the user directly but
        instead use Runner.run()

        Parameters
        ----------
        runner : Runner
            Parent runner for this run.
        command_obj : CommandManager
            CommandManager containing the subprocess executing this run.
        run_obj : Run
            Run object corresponding to this run.
        """
        self.runner = runner
        self.command_obj = command_obj

    def __enter__(self) -> "ExecutingProcess":
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.runner.__exit__(exc_type, exc_value, traceback)

    async def wait(
        self, timeout: Optional[float] = None, stream: Optional[str] = None
    ) -> "ExecutingProcess":
        """
        Wait for this run to finish, optionally with a timeout
        and optionally streaming its output.

        Note that this method is asynchronous and needs to be `await`ed.

        Parameters
        ----------
        timeout : float, optional, default None
            The maximum time, in seconds, to wait for the run to finish.
            If the timeout is reached, the run is terminated. If not specified, wait
            forever.
        stream : str, optional, default None
            If specified, the specified stream is printed to stdout. `stream` can
            be one of `stdout` or `stderr`.

        Returns
        -------
        ExecutingProcess
            This object, allowing you to chain calls.
        """
        await self.command_obj.wait(timeout, stream)
        return self

    @property
    def returncode(self) -> Optional[int]:
        """
        Gets the return code of the underlying subprocess. A non-zero
        code indicates a failure, `None` a currently executing run.

        Returns
        -------
        Optional[int]
            The return code of the underlying subprocess.
        """
        return self.command_obj.process.returncode

    @property
    def status(self) -> str:
        """
        Returns the status of the underlying subprocess that is responsible
        for executing the run.

        The return value is one of the following strings:
        - `timeout` indicates that the run timed out.
        - `running` indicates a currently executing run.
        - `failed` indicates a failed run.
        - `successful` indicates a successful run.

        Returns
        -------
        str
            The current status of the run.
        """
        if self.command_obj.timeout:
            return "timeout"
        elif self.command_obj.process.returncode is None:
            return "running"
        elif self.command_obj.process.returncode != 0:
            return "failed"
        else:
            return "successful"

    @property
    def stdout(self) -> str:
        """
        Returns the current stdout of the run. If the run is finished, this will
        contain the entire stdout output. Otherwise, it will contain the
        stdout up until this point.

        Returns
        -------
        str
            The current snapshot of stdout.
        """
        with open(
            self.command_obj.log_files.get("stdout"), "r", encoding="utf-8"
        ) as fp:
            return fp.read()

    @property
    def stderr(self) -> str:
        """
        Returns the current stderr of the run. If the run is finished, this will
        contain the entire stderr output. Otherwise, it will contain the
        stderr up until this point.

        Returns
        -------
        str
            The current snapshot of stderr.
        """
        with open(
            self.command_obj.log_files.get("stderr"), "r", encoding="utf-8"
        ) as fp:
            return fp.read()

    async def stream_log(
        self, stream: str, position: Optional[int] = None
    ) -> Iterator[Tuple[int, str]]:
        """
        Asynchronous iterator to stream logs from the subprocess line by line.

        Note that this method is asynchronous and needs to be `await`ed.

        Parameters
        ----------
        stream : str
            The stream to stream logs from. Can be one of `stdout` or `stderr`.
        position : int, optional, default None
            The position in the log file to start streaming from. If None, it starts
            from the beginning of the log file. This allows resuming streaming from
            a previously known position

        Yields
        ------
        Tuple[int, str]
            A tuple containing the position in the log file and the line read. The
            position returned can be used to feed into another `stream_logs` call
            for example.
        """
        async for position, line in self.command_obj.stream_log(stream, position):
            yield position, line


class ExecutingTask(ExecutingProcess):
    """
    This class contains a reference to a `metaflow.Task` object representing
    the currently executing or finished task, as well as metadata related
    to the process.
    `ExecutingTask` is returned by methods in `Runner` and `NBRunner`. It is not
    meant to be instantiated directly.
    This class works as a context manager, allowing you to use a pattern like
    ```python
    with Runner(...).spin() as running:
        ...
    ```
    Note that you should use either this object as the context manager or
    `Runner`, not both in a nested manner.
    """

    def __init__(
        self, runner: "Runner", command_obj: CommandManager, task_obj: Task
    ) -> None:
        """
        Create a new ExecutingTask -- this should not be done by the user directly but
        instead use Runner.spin()
        Parameters
        ----------
        runner : Runner
            Parent runner for this task.
        command_obj : CommandManager
            CommandManager containing the subprocess executing this task.
        task_obj : Task
            Task object corresponding to this task.
        """
        super().__init__(runner, command_obj)
        self.task = task_obj


class ExecutingRun(ExecutingProcess):
    """
    This class contains a reference to a `metaflow.Run` object representing
    the currently executing or finished run, as well as metadata related
    to the process.
    `ExecutingRun` is returned by methods in `Runner` and `NBRunner`. It is not
    meant to be instantiated directly.
    This class works as a context manager, allowing you to use a pattern like
    ```python
    with Runner(...).run() as running:
        ...
    ```
    Note that you should use either this object as the context manager or
    `Runner`, not both in a nested manner.
    """

    def __init__(
        self, runner: "Runner", command_obj: CommandManager, run_obj: Run
    ) -> None:
        """
        Create a new ExecutingRun -- this should not be done by the user directly but
        instead use Runner.run()
        Parameters
        ----------
        runner : Runner
            Parent runner for this run.
        command_obj : CommandManager
            CommandManager containing the subprocess executing this run.
        run_obj : Run
            Run object corresponding to this run.
        """
        super().__init__(runner, command_obj)
        self.run = run_obj


class RunnerMeta(type):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)

        def _injected_method(subcommand_name, runner_subcommand):
            def f(self, *args, **kwargs):
                return runner_subcommand(self, *args, **kwargs)

            f.__doc__ = runner_subcommand.__init__.__doc__ or ""
            f.__name__ = subcommand_name
            sig = inspect.signature(runner_subcommand)
            # We take all the same parameters except replace the first with
            # simple "self"
            new_parameters = {}
            for name, param in sig.parameters.items():
                if new_parameters:
                    new_parameters[name] = param
                else:
                    new_parameters["self"] = inspect.Parameter(
                        "self", inspect.Parameter.POSITIONAL_OR_KEYWORD
                    )
            f.__signature__ = inspect.Signature(
                list(new_parameters.values()), return_annotation=runner_subcommand
            )

            return f

        for runner_subcommand in get_runner_cli():
            method_name = runner_subcommand.name.replace("-", "_")
            setattr(cls, method_name, _injected_method(method_name, runner_subcommand))

        return cls


class Runner(metaclass=RunnerMeta):
    """
    Metaflow's Runner API that presents a programmatic interface
    to run flows and perform other operations either synchronously or asynchronously.
    The class expects a path to the flow file along with optional arguments
    that match top-level options on the command-line.

    This class works as a context manager, calling `cleanup()` to remove
    temporary files at exit.

    Example:
    ```python
    with Runner('slowflow.py', pylint=False) as runner:
        result = runner.run(alpha=5, tags=["abc", "def"], max_workers=5)
        print(result.run.finished)
    ```

    Parameters
    ----------
    flow_file : str
        Path to the flow file to run, relative to current directory.
    show_output : bool, default True
        Show the 'stdout' and 'stderr' to the console by default,
        Only applicable for synchronous 'run' and 'resume' functions.
    profile : str, optional, default None
        Metaflow profile to use to run this run. If not specified, the default
        profile is used (or the one already set using `METAFLOW_PROFILE`)
    env : Dict[str, str], optional, default None
        Additional environment variables to set for the Run. This overrides the
        environment set for this process.
    cwd : str, optional, default None
        The directory to run the subprocess in; if not specified, the current
        directory is used.
    file_read_timeout : int, default 3600
        The timeout until which we try to read the runner attribute file (in seconds).
    **kwargs : Any
        Additional arguments that you would pass to `python myflow.py` before
        the `run` command.
    """

    def __init__(
        self,
        flow_file: str,
        show_output: bool = True,
        profile: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        cwd: Optional[str] = None,
        file_read_timeout: int = 3600,
        **kwargs,
    ):
        # these imports are required here and not at the top
        # since they interfere with the user defined Parameters
        # in the flow file, this is related to the ability of
        # importing 'Runner' directly i.e.
        #    from metaflow import Runner
        # This ability is made possible by the statement:
        # 'from .metaflow_runner import Runner' in '__init__.py'

        from metaflow.parameters import flow_context

        # Reload the CLI with an "empty" flow -- this will remove any configuration
        # and parameter options. They are re-added in from_cli (called below).
        to_reload = [
            "metaflow.cli",
            "metaflow.cli_components.run_cmds",
            "metaflow.cli_components.init_cmd",
        ]
        with flow_context(None):
            [
                importlib.reload(sys.modules[module])
                for module in to_reload
                if module in sys.modules
            ]

        from metaflow.cli import start
        from metaflow.runner.click_api import MetaflowAPI

        # Convert flow_file to absolute path if it's relative
        if not os.path.isabs(flow_file):
            self.flow_file = os.path.abspath(flow_file)
        else:
            self.flow_file = flow_file

        self.show_output = show_output

        self.env_vars = os.environ.copy()
        self.env_vars.update(env or {})
        if profile:
            self.env_vars["METAFLOW_PROFILE"] = profile

        self.cwd = cwd or os.getcwd()
        self.file_read_timeout = file_read_timeout
        self.spm = SubprocessManager()
        self.top_level_kwargs = kwargs
        self.api = MetaflowAPI.from_cli(self.flow_file, start)

    def __enter__(self) -> "Runner":
        return self

    async def __aenter__(self) -> "Runner":
        return self

    def __get_executing_run(self, attribute_file_fd, command_obj):
        content = handle_timeout(attribute_file_fd, command_obj, self.file_read_timeout)

        command_obj.sync_wait()

        content = json.loads(content)
        pathspec = "%s/%s" % (content.get("flow_name"), content.get("run_id"))

        # Set the correct metadata from the runner_attribute file corresponding to this run.
        metadata_for_flow = content.get("metadata")

        run_object = Run(
            pathspec, _namespace_check=False, _current_metadata=metadata_for_flow
        )
        return ExecutingRun(self, command_obj, run_object)

    async def __async_get_executing_run(self, attribute_file_fd, command_obj):
        content = await async_handle_timeout(
            attribute_file_fd, command_obj, self.file_read_timeout
        )
        content = json.loads(content)
        pathspec = "%s/%s" % (content.get("flow_name"), content.get("run_id"))

        # Set the correct metadata from the runner_attribute file corresponding to this run.
        metadata_for_flow = content.get("metadata")

        run_object = Run(
            pathspec, _namespace_check=False, _current_metadata=metadata_for_flow
        )
        return ExecutingRun(self, command_obj, run_object)

    def run(self, **kwargs) -> ExecutingRun:
        """
        Blocking execution of the run. This method will wait until
        the run has completed execution.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments that you would pass to `python myflow.py` after
            the `run` command, in particular, any parameters accepted by the flow.

        Returns
        -------
        ExecutingRun
            ExecutingRun containing the results of the run.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            if CLICK_API_PROCESS_CONFIG:
                with with_dir(self.cwd):
                    command = self.api(**self.top_level_kwargs).run(
                        runner_attribute_file=attribute_file_path, **kwargs
                    )
            else:
                command = self.api(**self.top_level_kwargs).run(
                    runner_attribute_file=attribute_file_path, **kwargs
                )

            pid = self.spm.run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
                show_output=self.show_output,
            )
            command_obj = self.spm.get(pid)

            return self.__get_executing_run(attribute_file_fd, command_obj)

    def __get_executing_task(self, attribute_file_fd, command_obj):
        content = handle_timeout(attribute_file_fd, command_obj, self.file_read_timeout)

        command_obj.sync_wait()

        content = json.loads(content)
        pathspec = f"{content.get('flow_name')}/{content.get('run_id')}/{content.get('step_name')}/{content.get('task_id')}"

        # Set the correct metadata from the runner_attribute file corresponding to this run.
        metadata_for_flow = content.get("metadata")

        task_object = Task(
            pathspec, _namespace_check=False, _current_metadata=metadata_for_flow
        )
        return ExecutingTask(self, command_obj, task_object)

    async def __async_get_executing_task(self, attribute_file_fd, command_obj):
        content = await async_handle_timeout(
            attribute_file_fd, command_obj, self.file_read_timeout
        )
        content = json.loads(content)
        pathspec = f"{content.get('flow_name')}/{content.get('run_id')}/{content.get('step_name')}/{content.get('task_id')}"

        # Set the correct metadata from the runner_attribute file corresponding to this run.
        metadata_for_flow = content.get("metadata")

        task_object = Task(
            pathspec, _namespace_check=False, _current_metadata=metadata_for_flow
        )
        return ExecutingTask(self, command_obj, task_object)

    def spin(self, pathspec, **kwargs) -> ExecutingTask:
        """
        Blocking spin execution of the run.
        This method will wait until the spun run has completed execution.
        Parameters
        ----------
        pathspec : str
            The pathspec of the step/task to spin.
        **kwargs : Any
            Additional arguments that you would pass to `python ./myflow.py` after
            the `spin` command.
        Returns
        -------
        ExecutingTask
            ExecutingTask containing the results of the spun task.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            if CLICK_API_PROCESS_CONFIG:
                with with_dir(self.cwd):
                    command = self.api(**self.top_level_kwargs).spin(
                        pathspec=pathspec,
                        runner_attribute_file=attribute_file_path,
                        **kwargs,
                    )
            else:
                command = self.api(**self.top_level_kwargs).spin(
                    pathspec=pathspec,
                    runner_attribute_file=attribute_file_path,
                    **kwargs,
                )

            pid = self.spm.run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
                show_output=self.show_output,
            )
            command_obj = self.spm.get(pid)

            return self.__get_executing_task(attribute_file_fd, command_obj)

    def resume(self, **kwargs) -> ExecutingRun:
        """
        Blocking resume execution of the run.
        This method will wait until the resumed run has completed execution.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments that you would pass to `python ./myflow.py` after
            the `resume` command.

        Returns
        -------
        ExecutingRun
            ExecutingRun containing the results of the resumed run.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            if CLICK_API_PROCESS_CONFIG:
                with with_dir(self.cwd):
                    command = self.api(**self.top_level_kwargs).resume(
                        runner_attribute_file=attribute_file_path, **kwargs
                    )
            else:
                command = self.api(**self.top_level_kwargs).resume(
                    runner_attribute_file=attribute_file_path, **kwargs
                )

            pid = self.spm.run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
                show_output=self.show_output,
            )
            command_obj = self.spm.get(pid)

            return self.__get_executing_run(attribute_file_fd, command_obj)

    async def async_run(self, **kwargs) -> ExecutingRun:
        """
        Non-blocking execution of the run. This method will return as soon as the
        run has launched.

        Note that this method is asynchronous and needs to be `await`ed.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments that you would pass to `python myflow.py` after
            the `run` command, in particular, any parameters accepted by the flow.

        Returns
        -------
        ExecutingRun
            ExecutingRun representing the run that was started.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            if CLICK_API_PROCESS_CONFIG:
                with with_dir(self.cwd):
                    command = self.api(**self.top_level_kwargs).run(
                        runner_attribute_file=attribute_file_path, **kwargs
                    )
            else:
                command = self.api(**self.top_level_kwargs).run(
                    runner_attribute_file=attribute_file_path, **kwargs
                )

            pid = await self.spm.async_run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
            )
            command_obj = self.spm.get(pid)

            return await self.__async_get_executing_run(attribute_file_fd, command_obj)

    async def async_resume(self, **kwargs) -> ExecutingRun:
        """
        Non-blocking resume execution of the run.
        This method will return as soon as the resume has launched.

        Note that this method is asynchronous and needs to be `await`ed.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments that you would pass to `python myflow.py` after
            the `resume` command.

        Returns
        -------
        ExecutingRun
            ExecutingRun representing the resumed run that was started.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            if CLICK_API_PROCESS_CONFIG:
                with with_dir(self.cwd):
                    command = self.api(**self.top_level_kwargs).resume(
                        runner_attribute_file=attribute_file_path, **kwargs
                    )
            else:
                command = self.api(**self.top_level_kwargs).resume(
                    runner_attribute_file=attribute_file_path, **kwargs
                )

            pid = await self.spm.async_run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
            )
            command_obj = self.spm.get(pid)

            return await self.__async_get_executing_run(attribute_file_fd, command_obj)

    async def async_spin(self, pathspec, **kwargs) -> ExecutingTask:
        """
        Non-blocking spin execution of the run.
        This method will return as soon as the spun task has launched.

        Note that this method is asynchronous and needs to be `await`ed.

        Parameters
        ----------
        pathspec : str
            The pathspec of the step/task to spin.
        **kwargs : Any
            Additional arguments that you would pass to `python ./myflow.py` after
            the `spin` command.

        Returns
        -------
        ExecutingTask
            ExecutingTask representing the spun task that was started.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            if CLICK_API_PROCESS_CONFIG:
                with with_dir(self.cwd):
                    command = self.api(**self.top_level_kwargs).spin(
                        pathspec=pathspec,
                        runner_attribute_file=attribute_file_path,
                        **kwargs,
                    )
            else:
                command = self.api(**self.top_level_kwargs).spin(
                    pathspec=pathspec,
                    runner_attribute_file=attribute_file_path,
                    **kwargs,
                )

            pid = await self.spm.async_run_command(
                [sys.executable, *command],
                env=self.env_vars,
                cwd=self.cwd,
            )
            command_obj = self.spm.get(pid)

            return await self.__async_get_executing_task(attribute_file_fd, command_obj)

    def __exit__(self, exc_type, exc_value, traceback):
        self.spm.cleanup()

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.spm.cleanup()

    def cleanup(self):
        """
        Delete any temporary files created during execution.
        """
        self.spm.cleanup()



from typing import Union


class DummyFlow(object):
    def __init__(self, name="not_a_real_flow"):
        self.name = name


# This function is used to initialize the environment outside a flow.
def init_environment_outside_flow(
    flow: Union["metaflow.flowspec.FlowSpec", "metaflow.sidecar.DummyFlow"],
) -> "metaflow.metaflow_environment.MetaflowEnvironment":
    from metaflow.plugins import ENVIRONMENTS
    from metaflow.metaflow_config import DEFAULT_ENVIRONMENT
    from metaflow.metaflow_environment import MetaflowEnvironment

    return [
        e for e in ENVIRONMENTS + [MetaflowEnvironment] if e.TYPE == DEFAULT_ENVIRONMENT
    ][0](flow)



# Support saving of distribution information so we can give it back to users even
# if we do not install those distributions. This is used to package distributions in
# the MetaflowCodeContent package and provide an experience as if the packages were installed
# system-wide.

import os
import re
import sys
from pathlib import Path
from types import ModuleType
from typing import (
    Callable,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    TYPE_CHECKING,
    Union,
    cast,
)

import inspect
from collections import defaultdict

from ..extension_support import metadata
from ..util import get_metaflow_root

if TYPE_CHECKING:
    import pathlib

_cached_distributions = None

packages_distributions = None  # type: Optional[Callable[[], Mapping[str, List[str]]]]
name_normalizer = re.compile(r"[-_.]+")

if sys.version_info[:2] >= (3, 10):
    packages_distributions = metadata.packages_distributions
else:
    # This is the code present in 3.10+ -- we replicate here for other versions
    def _packages_distributions() -> Mapping[str, List[str]]:
        """
        Return a mapping of top-level packages to their
        distributions.
        """
        pkg_to_dist = defaultdict(list)
        for dist in metadata.distributions():
            for pkg in _top_level_declared(dist) or _top_level_inferred(dist):
                pkg_to_dist[pkg].append(dist.metadata["Name"])
        return dict(pkg_to_dist)

    def _top_level_declared(dist: metadata.Distribution) -> List[str]:
        return (dist.read_text("top_level.txt") or "").split()

    def _topmost(name: "pathlib.PurePosixPath") -> Optional[str]:
        """
        Return the top-most parent as long as there is a parent.
        """
        top, *rest = name.parts
        return top if rest else None

    def _get_toplevel_name(name: "pathlib.PurePosixPath") -> str:
        return _topmost(name) or (
            # python/typeshed#10328
            inspect.getmodulename(name)  # type: ignore
            or str(name)
        )

    def _top_level_inferred(dist: "metadata.Distribution"):
        opt_names = set(map(_get_toplevel_name, dist.files or []))

        def importable_name(name):
            return "." not in name

        return filter(importable_name, opt_names)

    packages_distributions = _packages_distributions


def modules_to_distributions() -> Dict[str, List[metadata.Distribution]]:
    """
    Return a mapping of top-level modules to their distributions.

    Returns
    -------
    Dict[str, List[metadata.Distribution]]
        A mapping of top-level modules to their distributions.
    """
    global _cached_distributions
    pd = cast(Callable[[], Mapping[str, List[str]]], packages_distributions)
    if _cached_distributions is None:
        _cached_distributions = {
            k: [metadata.distribution(d) for d in v] for k, v in pd().items()
        }
    return _cached_distributions


_ModuleInfo = NamedTuple(
    "_ModuleInfo",
    [
        ("name", str),
        ("root_paths", Set[str]),
        ("module", ModuleType),
        ("metaflow_module", bool),
    ],
)


class PackagedDistribution(metadata.Distribution):
    """
    A Python Package packaged within a MetaflowCodeContent. This allows users to use use importlib
    as they would regularly and the packaged Python Package would be considered as a
    distribution even if it really isn't (since it is just included in the PythonPath).
    """

    def __init__(self, root: str, content: Dict[str, str]):
        self._root = Path(root)
        self._content = content

    # Strongly inspired from PathDistribution in metadata.py
    def read_text(self, filename: Union[str, os.PathLike]) -> Optional[str]:
        if str(filename) in self._content:
            return self._content[str(filename)]
        return None

    read_text.__doc__ = metadata.Distribution.read_text.__doc__

    # Returns a metadata.SimplePath but not always present in importlib.metadata libs so
    # skipping return type.
    def locate_file(self, path: Union[str, os.PathLike]):
        return self._root / path


class PackagedDistributionFinder(metadata.DistributionFinder):
    def __init__(self, dist_info: Dict[str, Dict[str, str]]):
        self._dist_info = dist_info

    def find_distributions(self, context=metadata.DistributionFinder.Context()):
        if context.name is None:
            # Yields all known distributions
            for name, info in self._dist_info.items():
                yield PackagedDistribution(
                    os.path.join(get_metaflow_root(), name), info
                )
            return None
        name = name_normalizer.sub("-", cast(str, context.name)).lower()
        if name in self._dist_info:
            yield PackagedDistribution(
                os.path.join(get_metaflow_root(), cast(str, context.name)),
                self._dist_info[name],
            )
        return None



from metaflow import FlowSpec, step, IncludeFile, Parameter


def script_path(filename):
    """
    A convenience function to get the absolute path to a file in this
    tutorial's directory. This allows the tutorial to be launched from any
    directory.

    """
    import os

    filepath = os.path.join(os.path.dirname(__file__))
    return os.path.join(filepath, filename)


class PlayListFlow(FlowSpec):
    """
    A flow to help you build your favorite movie playlist.

    The flow performs the following steps:
    1) Ingests a CSV file containing metadata about movies.
    2) Loads two of the columns from the CSV into python lists.
    3) In parallel branches:
       - A) Filters movies by the genre parameter.
       - B) Choose a random movie from a different genre.
    4) Displays the top entries from the playlist.

    """

    movie_data = IncludeFile(
        "movie_data",
        help="The path to a movie metadata file.",
        default=script_path("movies.csv"),
    )

    genre = Parameter(
        "genre", help="Filter movies for a particular genre.", default="Sci-Fi"
    )

    recommendations = Parameter(
        "recommendations",
        help="The number of movies to recommend in " "the playlist.",
        default=5,
    )

    @step
    def start(self):
        """
        Parse the CSV file and load the values into a dictionary of lists.

        """
        import csv

        # For this example, we only need the movie title and the genres.
        columns = ["movie_title", "genres"]
        self.dataframe = {col: [] for col in columns}

        for row in csv.DictReader(self.movie_data.splitlines()):
            for col in columns:
                self.dataframe[col].append(row[col])

        # Compute genre-specific movies and a bonus movie in parallel.
        self.next(self.bonus_movie, self.genre_movies)

    @step
    def bonus_movie(self):
        """
        This step chooses a random movie from a different genre.

        """
        from random import choice

        # Find all the movies that are not in the provided genre.
        movies = [
            (movie, genres)
            for movie, genres in zip(
                self.dataframe["movie_title"], self.dataframe["genres"]
            )
            if self.genre.lower() not in genres.lower()
        ]

        # Choose one randomly.
        self.bonus = choice(movies)

        self.next(self.join)

    @step
    def genre_movies(self):
        """
        Filter the movies by genre.

        """
        from random import shuffle

        # Find all the movies titles in the specified genre.
        self.movies = [
            movie
            for movie, genres in zip(
                self.dataframe["movie_title"], self.dataframe["genres"]
            )
            if self.genre.lower() in genres.lower()
        ]

        # Randomize the title names.
        shuffle(self.movies)

        self.next(self.join)

    @step
    def join(self, inputs):
        """
        Join our parallel branches and merge results.

        """
        # Reassign relevant variables from our branches.
        self.playlist = inputs.genre_movies.movies
        self.bonus = inputs.bonus_movie.bonus

        self.next(self.end)

    @step
    def end(self):
        """
        Print out the playlist and bonus movie.

        """
        print("Playlist for movies in genre '%s'" % self.genre)
        for pick, movie in enumerate(self.playlist, start=1):
            print("Pick %d: '%s'" % (pick, movie))
            if pick >= self.recommendations:
                break

        print("Bonus Pick: '%s' from '%s'" % (self.bonus[0], self.bonus[1]))


if __name__ == "__main__":
    PlayListFlow()



from metaflow.metaflow_config import (
    DATASTORE_SPIN_LOCAL_DIR,
    DATASTORE_SYSROOT_SPIN,
)
from metaflow.plugins.datastores.local_storage import LocalStorage


class SpinStorage(LocalStorage):
    TYPE = "spin"
    METADATA_DIR = "_meta"
    DATASTORE_DIR = DATASTORE_SPIN_LOCAL_DIR  # ".metaflow_spin"
    SYSROOT_VAR = DATASTORE_SYSROOT_SPIN



import time
from .metadata_provider import MetaDatum


def clone_task_helper(
    flow_name,
    clone_run_id,
    run_id,
    step_name,
    clone_task_id,
    task_id,
    flow_datastore,
    metadata_service,
    origin_ds_set=None,
    attempt_id=0,
):
    # 1. initialize output datastore
    output = flow_datastore.get_task_datastore(
        run_id, step_name, task_id, attempt=attempt_id, mode="w"
    )
    output.init_task()

    origin_run_id, origin_step_name, origin_task_id = (
        clone_run_id,
        step_name,
        clone_task_id,
    )
    # 2. initialize origin datastore
    origin = None
    if origin_ds_set:
        origin = origin_ds_set.get_with_pathspec(
            "{}/{}/{}".format(origin_run_id, origin_step_name, origin_task_id)
        )
    else:
        origin = flow_datastore.get_task_datastore(
            origin_run_id, origin_step_name, origin_task_id
        )
    metadata_tags = ["attempt_id:{0}".format(attempt_id)]
    output.clone(origin)
    _ = metadata_service.register_task_id(
        run_id,
        step_name,
        task_id,
        attempt_id,
    )
    metadata_service.register_metadata(
        run_id,
        step_name,
        task_id,
        [
            MetaDatum(
                field="origin-task-id",
                value=str(origin_task_id),
                type="origin-task-id",
                tags=metadata_tags,
            ),
            MetaDatum(
                field="origin-run-id",
                value=str(origin_run_id),
                type="origin-run-id",
                tags=metadata_tags,
            ),
            MetaDatum(
                field="attempt",
                value=str(attempt_id),
                type="attempt",
                tags=metadata_tags,
            ),
            MetaDatum(
                field="attempt_ok",
                value="True",  # During clone, the task is always considered successful.
                type="internal_attempt_status",
                tags=metadata_tags,
            ),
        ],
    )
    output.done()



from metaflow import FlowSpec, step, IncludeFile


def script_path(filename):
    """
    A convenience function to get the absolute path to a file in this
    tutorial's directory. This allows the tutorial to be launched from any
    directory.

    """
    import os

    filepath = os.path.join(os.path.dirname(__file__))
    return os.path.join(filepath, filename)


class MovieStatsFlow(FlowSpec):
    """
    A flow to generate some statistics about the movie genres.

    The flow performs the following steps:
    1) Ingests a CSV into a dataframe.
    2) Fan-out over genre using Metaflow foreach.
    3) Compute quartiles for each genre.
    4) Save a dictionary of genre-specific statistics.

    """

    movie_data = IncludeFile(
        "movie_data",
        help="The path to a movie metadata file.",
        default=script_path("movies.csv"),
    )

    @step
    def start(self):
        """
        The start step:
        1) Loads the movie metadata into dataframe.
        2) Finds all the unique genres.
        3) Launches parallel statistics computation for each genre.

        """
        import csv

        # Load the dataset into a dataframe structure.
        columns = ["movie_title", "title_year", "genres", "gross"]
        self.dataframe = {col: [] for col in columns}

        for row in csv.DictReader(self.movie_data.splitlines()):
            for col in columns:
                val = int(row[col]) if col in ("title_year", "gross") else row[col]
                self.dataframe[col].append(val)

        # The column 'genres' has a list of genres for each movie. Let's get
        # all the unique genres.
        self.genres = {
            genre for genres in self.dataframe["genres"] for genre in genres.split("|")
        }
        self.genres = list(self.genres)

        # We want to compute some statistics for each genre. The 'foreach'
        # keyword argument allows us to compute the statistics for each genre in
        # parallel (i.e. a fan-out).
        self.next(self.compute_statistics, foreach="genres")

    @step
    def compute_statistics(self):
        """
        Compute statistics for a single genre.
        """

        # The genre currently being processed is a class property called
        # 'input'.
        self.genre = self.input
        print("Computing statistics for %s" % self.genre)

        # Find all the movies that have this genre and build a dataframe with
        # just those movies and just the columns of interest.
        selector = [self.genre in row for row in self.dataframe["genres"]]

        for col in self.dataframe.keys():
            self.dataframe[col] = [
                col for col, is_genre in zip(self.dataframe[col], selector) if is_genre
            ]

        # Sort by gross box office and drop unused column.
        argsort_indices = sorted(
            range(len(self.dataframe["gross"])), key=self.dataframe["gross"].__getitem__
        )
        for col in self.dataframe.keys():
            self.dataframe[col] = [self.dataframe[col][idx] for idx in argsort_indices]
        del self.dataframe["title_year"]

        # Get some statistics on the gross box office for these titles.
        n_points = len(self.dataframe["movie_title"])
        self.quartiles = []
        for cut in [0.25, 0.5, 0.75]:
            idx = 0 if n_points < 2 else round(n_points * cut)
            self.quartiles.append(self.dataframe["gross"][idx])

        # Join the results from other genres.
        self.next(self.join)

    @step
    def join(self, inputs):
        """
        Join our parallel branches and merge results into a dictionary.

        """
        # Merge results from the genre-specific computations.
        self.genre_stats = {
            inp.genre.lower(): {"quartiles": inp.quartiles, "dataframe": inp.dataframe}
            for inp in inputs
        }

        self.next(self.end)

    @step
    def end(self):
        """
        End the flow.

        """
        pass


if __name__ == "__main__":
    MovieStatsFlow()



from metaflow.decorators import FlowDecorator, StepDecorator
from metaflow.metaflow_environment import InvalidEnvironmentException


class PyPIStepDecorator(StepDecorator):
    """
    Specifies the PyPI packages for the step.

    Information in this decorator will augment any
    attributes set in the `@pyi_base` flow-level decorator. Hence,
    you can use `@pypi_base` to set packages required by all
    steps and use `@pypi` to specify step-specific overrides.

    Parameters
    ----------
    packages : Dict[str, str], default: {}
        Packages to use for this step. The key is the name of the package
        and the value is the version to use.
    python : str, optional, default: None
        Version of Python to use, e.g. '3.7.4'. A default value of None implies
        that the version used will correspond to the version of the Python interpreter used to start the run.
    """

    name = "pypi"
    defaults = {"packages": {}, "python": None, "disabled": None}  # wheels

    def __init__(self, attributes=None, statically_defined=False, inserted_by=None):
        self._attributes_with_user_values = (
            set(attributes.keys()) if attributes is not None else set()
        )

        super().__init__(attributes, statically_defined, inserted_by)

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        # The init_environment hook for Environment creates the relevant virtual
        # environments. The step_init hook sets up the relevant state for that hook to
        # do it's magic.

        self.flow = flow
        self.step = step

        # Support flow-level decorator
        if "pypi_base" in self.flow._flow_decorators:
            pypi_base = self.flow._flow_decorators["pypi_base"][0]
            super_attributes = pypi_base.attributes
            self._attributes_with_user_values.update(
                pypi_base._attributes_with_user_values
            )
            self.attributes["packages"] = {
                **super_attributes["packages"],
                **self.attributes["packages"],
            }
            self.attributes["python"] = (
                self.attributes["python"] or super_attributes["python"]
            )
            self.attributes["disabled"] = (
                self.attributes["disabled"]
                if self.attributes["disabled"] is not None
                else super_attributes["disabled"]
            )

        # Set default for `disabled` argument.
        if not self.attributes["disabled"]:
            self.attributes["disabled"] = False

        # At the moment, @pypi uses a conda environment as a virtual environment. This
        # is to ensure that we can have a dedicated Python interpreter within the
        # virtual environment. The conda environment is currently created through
        # micromamba. As a follow up, we can look into creating a virtualenv using
        # venv.

        # Currently, @pypi relies on pip for package resolution. We can introduce
        # support for Poetry in the near future, if desired. Poetry is great for
        # interactive use cases, but not so much for programmatic use cases like the
        # one here. We can consider introducing a UX where @pypi is able to consume
        # poetry.lock files in the future.

        _supported_virtual_envs = ["conda"]  # , "venv"]

        # To placate people who don't want to see a shred of conda in UX, we symlink
        # --environment=pypi to --environment=conda
        _supported_virtual_envs.extend(["pypi"])

        # TODO: Hardcoded for now to support the fast bakery environment.
        # We should introduce a more robust mechanism for appending supported environments, for example from within extensions.
        _supported_virtual_envs.extend(["fast-bakery"])

        # The --environment= requirement ensures that valid virtual environments are
        # created for every step to execute it, greatly simplifying the @pypi
        # implementation.
        if environment.TYPE not in _supported_virtual_envs:
            raise InvalidEnvironmentException(
                "@%s decorator requires %s"
                % (
                    self.name,
                    " or ".join(
                        ["--environment=%s" % env for env in _supported_virtual_envs]
                    ),
                )
            )
        # TODO: This code snippet can be done away with by altering the constructor of
        #       MetaflowEnvironment. A good first-task exercise.
        # Avoid circular import
        from metaflow.plugins.datastores.local_storage import LocalStorage

        environment.set_local_root(LocalStorage.get_datastore_root_from_config(logger))

    def is_attribute_user_defined(self, name):
        return name in self._attributes_with_user_values


class PyPIFlowDecorator(FlowDecorator):
    """
    Specifies the PyPI packages for all steps of the flow.

    Use `@pypi_base` to set common packages required by all
    steps and use `@pypi` to specify step-specific overrides.
    Parameters
    ----------
    packages : Dict[str, str], default: {}
        Packages to use for this flow. The key is the name of the package
        and the value is the version to use.
    python : str, optional, default: None
        Version of Python to use, e.g. '3.7.4'. A default value of None implies
        that the version used will correspond to the version of the Python interpreter used to start the run.
    """

    name = "pypi_base"
    defaults = {"packages": {}, "python": None, "disabled": None}

    def __init__(self, attributes=None, statically_defined=False, inserted_by=None):
        self._attributes_with_user_values = (
            set(attributes.keys()) if attributes is not None else set()
        )

        super().__init__(attributes, statically_defined, inserted_by)

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        from metaflow import decorators

        decorators._attach_decorators(flow, ["pypi"])
        decorators._process_late_attached_decorator(
            ["pypi"],
            flow,
            graph,
            environment,
            flow_datastore,
            logger,
        )

        # @pypi uses a conda environment to create a virtual environment.
        # The conda environment can be created through micromamba.
        _supported_virtual_envs = ["conda"]

        # To placate people who don't want to see a shred of conda in UX, we symlink
        # --environment=pypi to --environment=conda
        _supported_virtual_envs.extend(["pypi"])

        # TODO: Hardcoded for now to support the fast bakery environment.
        # We should introduce a more robust mechanism for appending supported environments, for example from within extensions.
        _supported_virtual_envs.extend(["fast-bakery"])

        # The --environment= requirement ensures that valid virtual environments are
        # created for every step to execute it, greatly simplifying the @conda
        # implementation.
        if environment.TYPE not in _supported_virtual_envs:
            raise InvalidEnvironmentException(
                "@%s decorator requires %s"
                % (
                    self.name,
                    " or ".join(
                        ["--environment=%s" % env for env in _supported_virtual_envs]
                    ),
                )
            )



from metaflow._vendor import click

from .. import namespace
from ..cli import echo_always, echo_dev_null
from ..cli_args import cli_args
from ..datastore.flow_datastore import FlowDataStore
from ..exception import CommandException
from ..client.filecache import FileCache, FileBlobCache, TaskMetadataCache
from ..metaflow_config import SPIN_ALLOWED_DECORATORS
from ..metaflow_profile import from_start
from ..plugins import DATASTORES
from ..task import MetaflowTask
from ..unbounded_foreach import UBF_CONTROL, UBF_TASK
from ..util import decompress_list, read_artifacts_module
import metaflow.tracing as tracing


@click.command(help="Internal command to execute a single task.", hidden=True)
@tracing.cli("cli/step")
@click.argument("step-name")
@click.option(
    "--run-id",
    default=None,
    required=True,
    help="ID for one execution of all steps in the flow.",
)
@click.option(
    "--task-id",
    default=None,
    required=True,
    show_default=True,
    help="ID for this instance of the step.",
)
@click.option(
    "--input-paths",
    help="A comma-separated list of pathspecs specifying inputs for this step.",
)
@click.option(
    "--input-paths-filename",
    type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
    help="A filename containing the argument typically passed to `input-paths`",
    hidden=True,
)
@click.option(
    "--split-index",
    type=int,
    default=None,
    show_default=True,
    help="Index of this foreach split.",
)
@click.option(
    "--tag",
    "opt_tag",
    multiple=True,
    default=None,
    help="Annotate this run with the given tag. You can specify "
    "this option multiple times to attach multiple tags in "
    "the task.",
)
@click.option(
    "--namespace",
    "opt_namespace",
    default=None,
    help="Change namespace from the default (your username) to the specified tag.",
)
@click.option(
    "--retry-count",
    default=0,
    help="How many times we have attempted to run this task.",
)
@click.option(
    "--max-user-code-retries",
    default=0,
    help="How many times we should attempt running the user code.",
)
@click.option(
    "--clone-only",
    default=None,
    help="Pathspec of the origin task for this task to clone. Do "
    "not execute anything.",
)
@click.option(
    "--clone-run-id",
    default=None,
    help="Run id of the origin flow, if this task is part of a flow being resumed.",
)
@click.option(
    "--ubf-context",
    default="none",
    type=click.Choice(["none", UBF_CONTROL, UBF_TASK]),
    help="Provides additional context if this task is of type unbounded foreach.",
)
@click.option(
    "--num-parallel",
    default=0,
    type=int,
    help="Number of parallel instances of a step. Ignored in local mode (see parallel decorator code).",
)
@click.pass_context
def step(
    ctx,
    step_name,
    opt_tag=None,
    run_id=None,
    task_id=None,
    input_paths=None,
    input_paths_filename=None,
    split_index=None,
    opt_namespace=None,
    retry_count=None,
    max_user_code_retries=None,
    clone_only=None,
    clone_run_id=None,
    ubf_context="none",
    num_parallel=None,
):
    if ctx.obj.is_quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    if ubf_context == "none":
        ubf_context = None
    if opt_namespace is not None:
        namespace(opt_namespace)

    func = None
    try:
        func = getattr(ctx.obj.flow, step_name)
    except:
        raise CommandException("Step *%s* doesn't exist." % step_name)
    if not func.is_step:
        raise CommandException("Function *%s* is not a step." % step_name)
    echo("Executing a step, *%s*" % step_name, fg="magenta", bold=False)

    step_kwargs = ctx.params
    # Remove argument `step_name` from `step_kwargs`.
    step_kwargs.pop("step_name", None)
    # Remove `opt_*` prefix from (some) option keys.
    step_kwargs = dict(
        [(k[4:], v) if k.startswith("opt_") else (k, v) for k, v in step_kwargs.items()]
    )
    cli_args._set_step_kwargs(step_kwargs)

    ctx.obj.metadata.add_sticky_tags(tags=opt_tag)
    if not input_paths and input_paths_filename:
        with open(input_paths_filename, mode="r", encoding="utf-8") as f:
            input_paths = f.read().strip(" \n\"'")

    paths = decompress_list(input_paths) if input_paths else []

    task = MetaflowTask(
        ctx.obj.flow,
        ctx.obj.flow_datastore,
        ctx.obj.metadata,
        ctx.obj.environment,
        ctx.obj.echo,
        ctx.obj.event_logger,
        ctx.obj.monitor,
        ubf_context,
    )
    if clone_only:
        task.clone_only(
            step_name,
            run_id,
            task_id,
            clone_only,
            retry_count,
        )
    else:
        task.run_step(
            step_name,
            run_id,
            task_id,
            clone_run_id,
            paths,
            split_index,
            retry_count,
            max_user_code_retries,
        )

    echo("Success", fg="green", bold=True, indent=True)


@click.command(help="Internal command to spin a single task.", hidden=True)
@click.argument("step-name")
@click.option(
    "--run-id",
    default=None,
    required=True,
    help="Original run ID for the step that will be spun",
)
@click.option(
    "--task-id",
    default=None,
    required=True,
    help="Original Task ID for the step that will be spun",
)
@click.option(
    "--orig-flow-datastore",
    show_default=True,
    help="Original datastore for the flow from which a task is being spun",
)
@click.option(
    "--input-paths",
    help="A comma-separated list of pathspecs specifying inputs for this step.",
)
@click.option(
    "--split-index",
    type=int,
    default=None,
    show_default=True,
    help="Index of this foreach split.",
)
@click.option(
    "--retry-count",
    default=0,
    help="How many times we have attempted to run this task.",
)
@click.option(
    "--max-user-code-retries",
    default=0,
    help="How many times we should attempt running the user code.",
)
@click.option(
    "--namespace",
    "opt_namespace",
    default=None,
    help="Change namespace from the default (your username) to the specified tag.",
)
@click.option(
    "--skip-decorators/--no-skip-decorators",
    is_flag=True,
    default=False,
    show_default=True,
    help="Skip decorators attached to the step or flow.",
)
@click.option(
    "--persist/--no-persist",
    "persist",
    default=True,
    show_default=True,
    help="Whether to persist the artifacts in the spun step. If set to false, the artifacts will not"
    " be persisted and will not be available in the spun step's datastore.",
)
@click.option(
    "--artifacts-module",
    default=None,
    show_default=True,
    help="Path to a module that contains artifacts to be used in the spun step. The artifacts should "
    "be defined as a dictionary called ARTIFACTS with keys as the artifact names and values as the "
    "artifact values. The artifact values will overwrite the default values of the artifacts used in "
    "the spun step.",
)
@click.pass_context
def spin_step(
    ctx,
    step_name,
    orig_flow_datastore,
    run_id=None,
    task_id=None,
    input_paths=None,
    split_index=None,
    retry_count=None,
    max_user_code_retries=None,
    opt_namespace=None,
    skip_decorators=False,
    artifacts_module=None,
    persist=True,
):
    import time

    if ctx.obj.is_quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    if opt_namespace is not None:
        namespace(opt_namespace)

    input_paths = decompress_list(input_paths) if input_paths else []

    skip_decorators = skip_decorators
    whitelist_decorators = [] if skip_decorators else SPIN_ALLOWED_DECORATORS
    from_start("SpinStep: initialized decorators")
    spin_artifacts = read_artifacts_module(artifacts_module) if artifacts_module else {}
    from_start("SpinStep: read artifacts module")

    ds_type, ds_root = orig_flow_datastore.split("@")
    orig_datastore_impl = [d for d in DATASTORES if d.TYPE == ds_type][0]
    orig_datastore_impl.datastore_root = ds_root
    orig_flow_datastore = FlowDataStore(
        ctx.obj.flow.name,
        environment=None,
        storage_impl=orig_datastore_impl,
        ds_root=ds_root,
    )

    filecache = FileCache()
    orig_flow_datastore.set_metadata_cache(
        TaskMetadataCache(filecache, ds_type, ds_root, ctx.obj.flow.name)
    )
    orig_flow_datastore.ca_store.set_blob_cache(
        FileBlobCache(
            filecache, FileCache.flow_ds_id(ds_type, ds_root, ctx.obj.flow.name)
        )
    )

    task = MetaflowTask(
        ctx.obj.flow,
        ctx.obj.flow_datastore,
        ctx.obj.metadata,
        ctx.obj.environment,
        echo,
        ctx.obj.event_logger,
        ctx.obj.monitor,
        None,  # no unbounded foreach context
        orig_flow_datastore=orig_flow_datastore,
        spin_artifacts=spin_artifacts,
    )
    from_start("SpinStep: initialized task")
    task.run_step(
        step_name,
        run_id,
        task_id,
        None,
        input_paths,
        split_index,
        retry_count,
        max_user_code_retries,
        whitelist_decorators,
        persist,
    )
    from_start("SpinStep: ran step")



import pickle

from metaflow._vendor import click

from ..cli import echo_always, echo_dev_null
from ..datastore import TaskDataStoreSet
from ..exception import CommandException


@click.command(
    help="Get data artifacts of a task or all tasks in a step. "
    "The format for input-path is either <run_id>/<step_name> or "
    "<run_id>/<step_name>/<task_id>."
)
@click.argument("input-path")
@click.option(
    "--private/--no-private",
    default=False,
    show_default=True,
    help="Show also private attributes.",
)
@click.option(
    "--max-value-size",
    default=1000,
    show_default=True,
    type=int,
    help="Show only values that are smaller than this number. "
    "Set to 0 to see only keys.",
)
@click.option(
    "--include",
    type=str,
    default="",
    help="Include only artifacts in the given comma-separated list.",
)
@click.option(
    "--file", type=str, default=None, help="Serialize artifacts in the given file."
)
@click.pass_obj
def dump(obj, input_path, private=None, max_value_size=None, include=None, file=None):

    if obj.is_quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    output = {}
    kwargs = {
        "show_private": private,
        "max_value_size": None if file is not None else max_value_size,
        "include": {t for t in include.split(",") if t},
    }

    # Pathspec can either be run_id/step_name or run_id/step_name/task_id.
    parts = input_path.split("/")
    if len(parts) == 2:
        run_id, step_name = parts
        task_id = None
    elif len(parts) == 3:
        run_id, step_name, task_id = parts
    else:
        raise CommandException(
            "input_path should either be run_id/step_name or run_id/step_name/task_id"
        )

    datastore_set = TaskDataStoreSet(
        obj.flow_datastore,
        run_id,
        steps=[step_name],
        prefetch_data_artifacts=kwargs.get("include"),
    )
    if task_id:
        ds_list = [datastore_set.get_with_pathspec(input_path)]
    else:
        ds_list = list(datastore_set)  # get all tasks

    for ds in ds_list:
        echo(
            "Dumping output of run_id=*{run_id}* "
            "step=*{step}* task_id=*{task_id}*".format(
                run_id=ds.run_id, step=ds.step_name, task_id=ds.task_id
            ),
            fg="magenta",
        )

        if file is None:
            echo_always(
                ds.format(**kwargs), highlight="green", highlight_bold=False, err=False
            )
        else:
            output[ds.pathspec] = ds.to_dict(**kwargs)

    if file is not None:
        with open(file, "wb") as f:
            pickle.dump(output, f, protocol=pickle.HIGHEST_PROTOCOL)
        echo("Artifacts written to *%s*" % file)
